from itertools import product

import narwhals as nw
import numpy as np
import pytest

from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
    WeightColumnFitMixinTests,
    WeightColumnInitMixinTests,
)
from tests.utils import assert_frame_equal_dispatch, dataframe_init_dispatch
from tubular.nominal import MeanResponseTransformer


# Dataframe used exclusively in this testing script
def create_MeanResponseTransformer_test_df(library="pandas"):
    """Create DataFrame to use MeanResponseTransformer tests that correct values are.

    DataFrame column a is the response, the other columns are categorical columns
    of types; object, category, int, float, bool.

    """
    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        "b": ["a", "b", "c", "d", "e", "f"],
        "c": ["a", "b", "c", "d", "e", "f"],
        "d": [1, 2, 3, 4, 5, 6],
        "e": [1, 2, 3, 4, 5, 6],
        "f": [False, False, False, True, True, True],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
        ],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


# Dataframe used exclusively in this testing script
def create_MeanResponseTransformer_test_df_unseen_levels(library="pandas"):
    """Create DataFrame to use in MeanResponseTransformer tests that check correct values are
    generated when using transform method on data with unseen levels.
    DataFrame column a is the response, the other columns are categorical columns
    of types; object, category, int, float, bool.

    """
    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 3.0],
        "b": ["a", "b", "c", "d", "e", "f", "g", "h"],
        "c": ["a", "b", "c", "d", "e", "f", "g", "h"],
        "d": [1, 2, 3, 4, 5, 6, 7, 8],
        "e": [1, 2, 3, 4, 5, 6, 7, 8],
        "f": [False, False, False, True, True, True, True, False],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
            "yellow",
            "blue",
        ],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)
    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


@pytest.fixture()
def learnt_mapping_dict():
    full_dict = {}

    b_dict = {
        "b": {"a": 1.0, "b": 2.0, "c": 3.0, "d": 4.0, "e": 5.0, "f": 6.0},
        "b_blue": {"a": 1.0, "b": 1.0, "c": 0.0, "d": 0.0, "e": 0.0, "f": 0.0},
        "b_yellow": {"a": 0.0, "b": 0.0, "c": 1.0, "d": 1.0, "e": 0.0, "f": 0.0},
        "b_green": {"a": 0.0, "b": 0.0, "c": 0.0, "d": 0.0, "e": 1.0, "f": 1.0},
    }

    # c matches b, but is categorical (see create_MeanResponseTransformer_test_df)
    c_dict = {
        "c" + suffix: b_dict["b" + suffix]
        for suffix in ["", "_blue", "_yellow", "_green"]
    }

    full_dict.update(b_dict)
    full_dict.update(c_dict)

    return full_dict


@pytest.fixture()
def learnt_unseen_levels_encoding_dict_mean():
    return_dict = {
        "b": (1.0 + 2.0 + 3.0 + 4.0 + 5.0 + 6.0) / 6,
        "b_blue": (1.0 + 1.0 + 0.0 + 0.0 + 0.0 + 0.0) / 6,
        "b_yellow": (0.0 + 0.0 + 1.0 + 1.0 + 0.0 + 0.0) / 6,
        "b_green": (0.0 + 0.0 + 0.0 + 0.0 + 1.0 + 1.0) / 6,
    }

    for key in return_dict:
        return_dict[key] = np.float32(return_dict[key])
    return return_dict


@pytest.fixture()
def learnt_unseen_levels_encoding_dict_median():
    return_dict = {
        "b": 3.0,
        "b_blue": 0.0,
        "b_yellow": 0.0,
        "b_green": 0.0,
    }

    for key in return_dict:
        return_dict[key] = np.float32(return_dict[key])
    return return_dict


@pytest.fixture()
def learnt_unseen_levels_encoding_dict_highest():
    return_dict = {
        "b": 6.0,
        "b_blue": 1.0,
        "b_yellow": 1.0,
        "b_green": 1.0,
    }
    for key in return_dict:
        return_dict[key] = np.float32(return_dict[key])
    return return_dict


@pytest.fixture()
def learnt_unseen_levels_encoding_dict_lowest():
    return_dict = {
        "b": 1.0,
        "b_blue": 0.0,
        "b_yellow": 0.0,
        "b_green": 0.0,
    }

    for key in return_dict:
        return_dict[key] = np.float32(return_dict[key])
    return return_dict


@pytest.fixture()
def learnt_unseen_levels_encoding_dict_arbitrary():
    return_dict = {
        "b": 22.0,
        "b_blue": 22.0,
        "b_yellow": 22.0,
        "b_green": 22.0,
    }
    for key in return_dict:
        return_dict[key] = np.float32(return_dict[key])
    return return_dict


class TestInit(ColumnStrListInitTests, WeightColumnInitMixinTests):
    """Tests for MeanResponseTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MeanResponseTransformer"

    # overload inherited arg tests that have been replaced by beartype

    def test_weight_arg_errors(self):
        pass

    def test_prior_not_positive_int_error(self):
        """Test that an exception is raised if prior is not a positive int."""
        with pytest.raises(ValueError, match="prior should be positive int"):
            MeanResponseTransformer(prior=-1)


class TestPriorRegularisation:
    "tests for _prior_regularisation method."

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_output1(self, library):
        "Test output of method."
        x = MeanResponseTransformer(columns="a", prior=3)

        df_dict = {"a": [1, 2]}
        df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)
        y = nw.new_series(name="y", values=[2, 3], backend=library)

        x.fit(X=df, y=y)

        expected1 = (1 + 3 * 2.5) / (1 + 3)

        expected2 = (2 + 3 * 2.5) / (2 + 3)

        expected = {"a": expected1, "b": expected2}

        weights_column = "weights"
        response_column = "means"
        column = "column"
        group_means_and_weights_dict = {
            column: ["a", "b"],
            response_column: [1, 2],
            weights_column: [1, 2],
        }

        group_means_and_weights_df = dataframe_init_dispatch(
            dataframe_dict=group_means_and_weights_dict,
            library="pandas",
        )

        output = x._prior_regularisation(
            group_means_and_weights=group_means_and_weights_df,
            column=column,
            weights_column=weights_column,
            response_column=response_column,
        )

        assert (
            output == expected
        ), f"output of _prior_regularisation not as expected, expected {expected} but got {output}"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_output2(self, library):
        "Test output of method - for category dtypes"
        x = MeanResponseTransformer(columns="a", prior=0)

        df_dict = {"a": ["a", "b"]}
        df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

        df = nw.from_native(df)
        df = df.with_columns(nw.col("a").cast(nw.Categorical))
        df = nw.to_native(df)

        y = nw.new_series(name="y", values=[2, 3], backend=library)

        x.fit(X=df, y=y)

        expected1 = (1) / (1)

        expected2 = (2) / (2)

        expected = {"a": expected1, "b": expected2}

        weights_column = "weights"
        response_column = "means"
        column = "column"
        group_means_and_weights_dict = {
            column: ["a", "b"],
            response_column: [1, 2],
            weights_column: [1, 2],
        }

        group_means_and_weights_df = dataframe_init_dispatch(
            dataframe_dict=group_means_and_weights_dict,
            library="pandas",
        )

        output = x._prior_regularisation(
            group_means_and_weights=group_means_and_weights_df,
            column=column,
            weights_column=weights_column,
            response_column=response_column,
        )

        assert (
            output == expected
        ), f"output of _prior_regularisation not as expected, expected {expected} but got {output}"

    @pytest.mark.parametrize("library", ["pandas"])
    def test_output3(self, library):
        "Test output of method - for pandas object dtype"
        x = MeanResponseTransformer(columns="a", prior=0)

        df_dict = {"a": ["a", "b"]}
        df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

        df["a"] = df["a"].astype("object")

        y = nw.new_series(name="y", values=[2, 3], backend=library)

        x.fit(X=df, y=y)

        expected1 = (1) / (1)

        expected2 = (2) / (2)

        expected = {"a": expected1, "b": expected2}

        weights_column = "weights"
        response_column = "means"
        column = "column"
        group_means_and_weights_dict = {
            column: ["a", "b"],
            response_column: [1, 2],
            weights_column: [1, 2],
        }

        group_means_and_weights_df = dataframe_init_dispatch(
            dataframe_dict=group_means_and_weights_dict,
            library="pandas",
        )

        output = x._prior_regularisation(
            group_means_and_weights=group_means_and_weights_df,
            column=column,
            weights_column=weights_column,
            response_column=response_column,
        )

        assert (
            output == expected
        ), f"output of _prior_regularisation not as expected, expected {expected} but got {output}"


class TestFit(GenericFitTests, WeightColumnFitMixinTests):
    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MeanResponseTransformer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_weights_column_missing_error(self, library):
        """Test that an exception is raised if weights_column is specified but not present in data for fit."""
        df = create_MeanResponseTransformer_test_df(library=library)

        x = MeanResponseTransformer(weights_column="z", columns=["b", "d", "f"])

        with pytest.raises(
            ValueError,
            match=r"weight col \(z\) is not present in columns of data",
        ):
            x.fit(
                df,
                df["a"],
            )

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("level", "target_column", "unseen_level_handling"),
        [
            (None, "a", "mean"),
            ("all", "multi_level_response", 32),
            (["yellow", "blue"], "multi_level_response", "max"),
        ],
    )
    def test_response_column_nulls_error(
        self,
        level,
        target_column,
        unseen_level_handling,
        library,
    ):
        """Test that an exception is raised if nulls are present in response_column."""
        df = create_MeanResponseTransformer_test_df(library=library)

        df = nw.from_native(df)
        df = df.with_columns(
            nw.new_series(
                name=target_column,
                values=[*[1.0] * (len(df) - 1), None],
                backend=library,
            ),
        )
        df = nw.to_native(df)

        x = MeanResponseTransformer(
            columns=["b"],
            level=level,
            unseen_level_handling=unseen_level_handling,
        )

        with pytest.raises(
            ValueError,
            match="MeanResponseTransformer: y has 1 null values",
        ):
            x.fit(df, df[target_column])

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("level", "target_column", "unseen_level_handling"),
        [
            (None, "a", "mean"),
            (None, "a", "min"),
        ],
    )
    def test_correct_mappings_stored_numeric_response(
        self,
        learnt_mapping_dict,
        level,
        target_column,
        unseen_level_handling,
        library,
    ):
        "Test that the mapping dictionary created in fit has the correct keys and values."
        df = create_MeanResponseTransformer_test_df(library=library)
        columns = ["b", "c"]
        x = MeanResponseTransformer(
            columns=columns,
            level=level,
            unseen_level_handling=unseen_level_handling,
        )
        x.fit(df, df[target_column])

        assert x.columns == columns, "Columns attribute changed in fit"

        for column in x.columns:
            actual = x.mappings[column]
            expected = learnt_mapping_dict[column]
            assert actual == expected

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("level", "target_column", "unseen_level_handling"),
        [
            (["blue"], "multi_level_response", "median"),
            ("all", "multi_level_response", 32),
            (["yellow", "blue"], "multi_level_response", "max"),
        ],
    )
    def test_correct_mappings_stored_categorical_response(
        self,
        learnt_mapping_dict,
        level,
        target_column,
        unseen_level_handling,
        library,
    ):
        "Test that the mapping dictionary created in fit has the correct keys and values."
        df = create_MeanResponseTransformer_test_df(library=library)
        columns = ["b", "c"]
        x = MeanResponseTransformer(
            columns=columns,
            level=level,
            unseen_level_handling=unseen_level_handling,
        )
        x.fit(df, df[target_column])

        df = nw.from_native(df)

        if level == "all":
            expected_created_cols = {
                prefix + "_" + suffix
                for prefix, suffix in product(
                    columns,
                    df[target_column].unique().to_list(),
                )
            }

        else:
            expected_created_cols = {
                prefix + "_" + suffix for prefix, suffix in product(columns, level)
            }
        assert (
            set(x.encoded_columns) == expected_created_cols
        ), "Stored encoded columns are not as expected"

        for column in x.encoded_columns:
            actual = x.mappings[column]
            expected = learnt_mapping_dict[column]
            assert actual == expected

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("level", "target_column", "unseen_level_handling"),
        [
            (None, "a", "mean"),
            (None, "a", "median"),
            (None, "a", "min"),
            (None, "a", "max"),
            (None, "a", 22.0),
            ("all", "multi_level_response", "mean"),
            (["yellow", "blue"], "multi_level_response", "mean"),
        ],
    )
    def test_correct_unseen_levels_encoding_dict_stored(
        self,
        learnt_unseen_levels_encoding_dict_mean,
        learnt_unseen_levels_encoding_dict_median,
        learnt_unseen_levels_encoding_dict_lowest,
        learnt_unseen_levels_encoding_dict_highest,
        learnt_unseen_levels_encoding_dict_arbitrary,
        level,
        target_column,
        unseen_level_handling,
        library,
    ):
        "Test that the unseen_levels_encoding_dict dictionary created in fit has the correct keys and values."
        df = create_MeanResponseTransformer_test_df(library=library)
        x = MeanResponseTransformer(
            columns=["b"],
            level=level,
            unseen_level_handling=unseen_level_handling,
        )
        x.fit(df, df[target_column])

        if level:
            if level == "all":
                assert set(x.unseen_levels_encoding_dict.keys()) == {
                    "b_blue",
                    "b_yellow",
                    "b_green",
                }, "Stored unseen_levels_encoding_dict keys are not as expected"

            else:
                assert set(x.unseen_levels_encoding_dict.keys()) == {
                    "b_blue",
                    "b_yellow",
                }, "Stored unseen_levels_encoding_dict keys are not as expected"

            for column in x.unseen_levels_encoding_dict:
                actual = x.unseen_levels_encoding_dict[column]
                expected = learnt_unseen_levels_encoding_dict_mean[column]
                assert actual == expected

        else:
            assert x.unseen_levels_encoding_dict.keys() == {
                "b",
            }, "Stored unseen_levels_encoding_dict key is not as expected"

            if unseen_level_handling == "mean":
                for column in x.unseen_levels_encoding_dict:
                    actual = x.unseen_levels_encoding_dict[column]
                    expected = learnt_unseen_levels_encoding_dict_mean[column]
                    assert actual == expected

            elif unseen_level_handling == "median":
                for column in x.unseen_levels_encoding_dict:
                    actual = x.unseen_levels_encoding_dict[column]
                    expected = learnt_unseen_levels_encoding_dict_median[column]
                    assert actual == expected

            elif unseen_level_handling == "min":
                for column in x.unseen_levels_encoding_dict:
                    actual = x.unseen_levels_encoding_dict[column]
                    expected = learnt_unseen_levels_encoding_dict_lowest[column]
                    assert actual == expected

            elif unseen_level_handling == "max":
                for column in x.unseen_levels_encoding_dict:
                    actual = x.unseen_levels_encoding_dict[column]
                    expected = learnt_unseen_levels_encoding_dict_highest[column]
                    assert actual == expected

            else:
                for column in x.unseen_levels_encoding_dict:
                    actual = x.unseen_levels_encoding_dict[column]
                    expected = learnt_unseen_levels_encoding_dict_arbitrary[column]
                    assert actual == expected

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_missing_categories_ignored(self, library):
        "test that where a categorical column has missing levels, these do not make it into the encoding dict"

        df = create_MeanResponseTransformer_test_df(library=library)
        unobserved_value = "a"
        df = nw.from_native(df)
        df = df.filter(nw.col("c") == unobserved_value)
        df = nw.to_native(df)

        target_column = "e"
        x = MeanResponseTransformer(
            columns=["c"],
        )
        x.fit(df, df[target_column])

        assert (
            unobserved_value not in x.mappings
        ), "MeanResponseTransformer should ignore unobserved levels"


class TestFitBinaryResponse(GenericFitTests, WeightColumnFitMixinTests):
    """Tests for MeanResponseTransformer.fit()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MeanResponseTransformer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        (
            "columns",
            "weights_values",
            "prior",
            "expected_mappings",
            "expected_mean",
        ),
        [
            # no prior, no weight
            (
                ["b", "d", "f"],
                [1, 1, 1, 1, 1, 1],
                0,
                {
                    "b": {"a": 1.0, "b": 2.0, "c": 3.0, "d": 4.0, "e": 5.0, "f": 6.0},
                    "d": {1: 1.0, 2: 2.0, 3: 3.0, 4: 4.0, 5: 5.0, 6: 6.0},
                    "f": {False: 2.0, True: 5.0},
                },
                np.float64(3.5),
            ),
            # no weight, prior
            (
                ["b", "d", "f"],
                [1, 1, 1, 1, 1, 1],
                5,
                {
                    "b": {
                        "a": 37 / 12,
                        "b": 13 / 4,
                        "c": 41 / 12,
                        "d": 43 / 12,
                        "e": 15 / 4,
                        "f": 47 / 12,
                    },
                    "d": {
                        1: 37 / 12,
                        2: 13 / 4,
                        3: 41 / 12,
                        4: 43 / 12,
                        5: 15 / 4,
                        6: 47 / 12,
                    },
                    "f": {False: 47 / 16, True: 65 / 16},
                },
                np.float64(3.5),
            ),
            # weight, no prior
            (
                ["b", "d", "f"],
                [1, 2, 3, 4, 5, 6],
                0,
                {
                    "b": {"a": 1.0, "b": 2.0, "c": 3.0, "d": 4.0, "e": 5.0, "f": 6.0},
                    "d": {1: 1.0, 2: 2.0, 3: 3.0, 4: 4.0, 5: 5.0, 6: 6.0},
                    "f": {False: 14 / 6, True: 77 / 15},
                },
                np.float64(13 / 3),
            ),
            # prior and weight
            (
                ["d", "f"],
                [1, 1, 1, 2, 2, 2],
                5,
                {
                    "d": {1: 7 / 2, 2: 11 / 3, 3: 23 / 6, 4: 4.0, 5: 30 / 7, 6: 32 / 7},
                    "f": {False: 13 / 4, True: 50 / 11},
                },
                np.float64(4.0),
            ),
        ],
    )
    def test_learnt_values(
        self,
        library,
        columns,
        weights_values,
        prior,
        expected_mappings,
        expected_mean,
    ):
        """Test that the mean response values learnt during fit are expected."""
        df = create_MeanResponseTransformer_test_df(library=library)

        df = nw.from_native(df)

        weights_column = "weights_column"
        df = df.with_columns(
            nw.new_series(name=weights_column, values=weights_values, backend=library),
        ).to_native()

        x = MeanResponseTransformer(columns=columns, prior=prior)

        x.mappings = {}

        x._fit_binary_response(
            df,
            x.columns,
            weights_column=weights_column,
            response_column="a",
        )

        for key in expected_mappings:
            for value in expected_mappings[key]:
                expected_mappings[key][value] = x.cast_method(
                    expected_mappings[key][value],
                )

        assert (
            x.global_mean == expected_mean
        ), f"global mean not learnt as expected, expected {expected_mean} but got {x.global_mean}"

        assert (
            x.mappings == expected_mappings
        ), f"mappings not learnt as expected, expected {expected_mappings} but got {x.mappings}"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize("prior", (1, 3, 5, 7, 9, 11, 100))
    def test_prior_logic(self, prior, library):
        "Test that for prior>0 encodings are closer to global mean than for prior=0."
        df = create_MeanResponseTransformer_test_df(library=library)

        df = nw.from_native(df)
        weights_column = "weights_column"
        native_backend = nw.get_native_namespace(df)
        df = df.with_columns(
            nw.new_series(
                name=weights_column,
                values=[1, 1, 1, 2, 2, 2],
                backend=native_backend.__name__,
            ),
        )

        x_prior = MeanResponseTransformer(
            columns=["d", "f"],
            prior=prior,
            weights_column=weights_column,
        )

        x_no_prior = MeanResponseTransformer(
            columns=["d", "f"],
            prior=0,
            weights_column=weights_column,
        )

        x_prior.mappings = {}
        x_no_prior.mappings = {}

        x_prior._fit_binary_response(
            df,
            x_prior.columns,
            response_column="a",
            weights_column=weights_column,
        )

        x_no_prior._fit_binary_response(
            df,
            x_no_prior.columns,
            weights_column=weights_column,
            response_column="a",
        )

        prior_mappings = x_prior.mappings

        no_prior_mappings = x_no_prior.mappings

        global_mean = x_prior.global_mean

        assert (
            global_mean == x_no_prior.global_mean
        ), "global means for transformers with/without priors should match"

        for col in prior_mappings:
            for value in prior_mappings[col]:
                prior_encoding = prior_mappings[col][value]
                no_prior_encoding = no_prior_mappings[col][value]

                prior_mean_dist = np.abs(prior_encoding - global_mean)
                no_prior_mean_dist = np.abs(no_prior_encoding - global_mean)

                assert (
                    prior_mean_dist <= no_prior_mean_dist
                ), "encodings using priors should be closer to the global mean than without"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("low_weight", "high_weight"),
        ((1, 2), (2, 3), (3, 4), (10, 20)),
    )
    def test_prior_logic_for_weights(self, low_weight, high_weight, library):
        "Test that for fixed prior a group with lower weight is moved closer to the global mean than one with higher weight."
        df = create_MeanResponseTransformer_test_df(library=library)

        df = nw.from_native(df)
        weights_column = "weights_column"
        native_backend = nw.get_native_namespace(df)

        # column f looks like [False, False, False, True, True, True]
        df = df.with_columns(
            nw.new_series(
                name=weights_column,
                values=[
                    low_weight,
                    low_weight,
                    low_weight,
                    high_weight,
                    high_weight,
                    high_weight,
                ],
                backend=native_backend.__name__,
            ),
        )

        x_prior = MeanResponseTransformer(
            columns=["f"],
            prior=5,
            weights_column=weights_column,
        )

        x_no_prior = MeanResponseTransformer(
            columns=["f"],
            prior=0,
            weights_column=weights_column,
        )

        x_prior.mappings = {}
        x_no_prior.mappings = {}

        x_prior._fit_binary_response(
            df,
            x_prior.columns,
            weights_column=weights_column,
            response_column="a",
        )

        x_no_prior._fit_binary_response(
            df,
            x_no_prior.columns,
            weights_column=weights_column,
            response_column="a",
        )

        prior_mappings = x_prior.mappings

        no_prior_mappings = x_no_prior.mappings

        global_mean = x_prior.global_mean

        assert (
            global_mean == x_no_prior.global_mean
        ), "global means for transformers with/without priors should match"

        low_weight_prior_encoding = prior_mappings["f"][False]
        high_weight_prior_encoding = prior_mappings["f"][True]

        low_weight_no_prior_encoding = no_prior_mappings["f"][False]
        high_weight_no_prior_encoding = no_prior_mappings["f"][True]

        low_weight_prior_mean_dist = np.abs(low_weight_prior_encoding - global_mean)
        high_weight_prior_mean_dist = np.abs(high_weight_prior_encoding - global_mean)

        low_weight_no_prior_mean_dist = np.abs(
            low_weight_no_prior_encoding - global_mean,
        )
        high_weight_no_prior_mean_dist = np.abs(
            high_weight_no_prior_encoding - global_mean,
        )

        # check low weight group has been moved further towards mean than high weight group by prior, i.e
        # that the distance remaining is a smaller proportion of the no prior distance
        low_ratio = low_weight_prior_mean_dist / low_weight_no_prior_mean_dist
        high_ratio = high_weight_prior_mean_dist / high_weight_no_prior_mean_dist
        assert (
            low_ratio <= high_ratio
        ), "encodings for categories with lower weights should be moved closer to the global mean than those with higher weights, for fixed prior"


def expected_df_1(library="pandas"):
    """Expected output for single level response."""

    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        "b": [1, 2, 3, 4, 5, 6],
        "c": ["a", "b", "c", "d", "e", "f"],
        "d": [1, 2, 3, 4, 5, 6],
        "e": [1, 2, 3, 4, 5, 6],
        "f": [2, 2, 2, 5, 5, 5],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
        ],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


def expected_df_2(library="pandas"):
    """Expected output for response with level = blue."""

    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        "c": ["a", "b", "c", "d", "e", "f"],
        "d": [1, 2, 3, 4, 5, 6],
        "e": [1, 2, 3, 4, 5, 6],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
        ],
        "b_blue": [1, 1, 0, 0, 0, 0],
        "f_blue": [2 / 3, 2 / 3, 2 / 3, 0, 0, 0],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


def expected_df_3(library="pandas"):
    """Expected output for response with level = 'all'."""

    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        "c": ["a", "b", "c", "d", "e", "f"],
        "d": [1, 2, 3, 4, 5, 6],
        "e": [1, 2, 3, 4, 5, 6],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
        ],
        "b_blue": [1, 1, 0, 0, 0, 0],
        "f_blue": [2 / 3, 2 / 3, 2 / 3, 0.0, 0.0, 0.0],
        "b_green": [0, 0, 0, 0, 1, 1],
        "f_green": [0.0, 0.0, 0.0, 2 / 3, 2 / 3, 2 / 3],
        "b_yellow": [0, 0, 1, 1, 0, 0],
        "f_yellow": [1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


def expected_df_4(library="pandas"):
    """Expected output for transform on dataframe with single level response and unseen levels,
    where unseen_level_handling = 'mean'.
    """
    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 3.0],
        "b": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 3.5, 3.5],
        "c": ["a", "b", "c", "d", "e", "f", "g", "h"],
        "d": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 3.5, 3.5],
        "e": [1, 2, 3, 4, 5, 6, 7, 8],
        "f": [2.0, 2.0, 2.0, 5.0, 5.0, 5.0, 5.0, 2.0],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
            "yellow",
            "blue",
        ],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)
    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


def expected_df_5(library="pandas"):
    """Expected output for transform on dataframe with single level response and unseen levels,
    where unseen_level_handling = 'median'.
    """

    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 3.0],
        "b": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 3.0, 3.0],
        "c": ["a", "b", "c", "d", "e", "f", "g", "h"],
        "d": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 3.0, 3.0],
        "e": [1, 2, 3, 4, 5, 6, 7, 8],
        "f": [2.0, 2.0, 2.0, 5.0, 5.0, 5.0, 5.0, 2.0],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
            "yellow",
            "blue",
        ],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)
    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


def expected_df_6(library="pandas"):
    """Expected output for transform on dataframe with single level response and unseen levels,
    where unseen_level_handling = 'min'.
    """
    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 3.0],
        "b": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 1.0, 1.0],
        "c": ["a", "b", "c", "d", "e", "f", "g", "h"],
        "d": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 1.0, 1.0],
        "e": [1, 2, 3, 4, 5, 6, 7, 8],
        "f": [2.0, 2.0, 2.0, 5.0, 5.0, 5.0, 5.0, 2.0],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
            "yellow",
            "blue",
        ],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)
    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


def expected_df_7(library="pandas"):
    """Expected output for transform on dataframe with single level response and unseen levels,
    where unseen_level_handling = 'max'.
    """
    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 3.0],
        "b": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 6.0, 6.0],
        "c": ["a", "b", "c", "d", "e", "f", "g", "h"],
        "d": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 6.0, 6.0],
        "e": [1, 2, 3, 4, 5, 6, 7, 8],
        "f": [2.0, 2.0, 2.0, 5.0, 5.0, 5.0, 5.0, 2.0],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
            "yellow",
            "blue",
        ],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)
    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


def expected_df_8(library="pandas"):
    """Expected output for transform on dataframe with single level response and unseen levels,
    where unseen_level_handling set to arbitrary int/float value.
    """
    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 3.0],
        "b": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 21.6, 21.6],
        "c": ["a", "b", "c", "d", "e", "f", "g", "h"],
        "d": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 21.6, 21.6],
        "e": [1, 2, 3, 4, 5, 6, 7, 8],
        "f": [2.0, 2.0, 2.0, 5.0, 5.0, 5.0, 5.0, 2.0],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
            "yellow",
            "blue",
        ],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)
    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


def expected_df_9(library="pandas"):
    """Expected output for transform on dataframe with multi-level response with level = blue and unseen levels in data."""
    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 3.0],
        "c": ["a", "b", "c", "d", "e", "f", "g", "h"],
        "d": [1, 2, 3, 4, 5, 6, 7, 8],
        "e": [1, 2, 3, 4, 5, 6, 7, 8],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
            "yellow",
            "blue",
        ],
        "b_blue": [1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 2 / 6, 2 / 6],
        "f_blue": [2 / 3, 2 / 3, 2 / 3, 0.0, 0.0, 0.0, 0, 2 / 3],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)
    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


def expected_df_10(library="pandas"):
    """Expected output for transform on dataframe with multi-level response with level = "all" and unseen levels in data."""
    df_dict = {
        "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 3.0],
        "c": ["a", "b", "c", "d", "e", "f", "g", "h"],
        "d": [1, 2, 3, 4, 5, 6, 7, 8],
        "e": [1, 2, 3, 4, 5, 6, 7, 8],
        "multi_level_response": [
            "blue",
            "blue",
            "yellow",
            "yellow",
            "green",
            "green",
            "yellow",
            "blue",
        ],
        "b_blue": [1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0],
        "f_blue": [2 / 3, 2 / 3, 2 / 3, 0, 0, 0, 0, 2 / 3],
        "b_yellow": [0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0],
        "f_yellow": [1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3],
        "b_green": [0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0],
        "f_green": [0.0, 0.0, 0.0, 2 / 3, 2 / 3, 2 / 3, 2 / 3, 0],
    }

    df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)
    df = nw.from_native(df)
    df = df.with_columns(nw.col("c").cast(nw.Categorical))

    return df.to_native()


class TestTransform(GenericTransformTests):
    """Tests for MeanResponseTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MeanResponseTransformer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        (
            "columns",
            "level",
            "mappings",
            "column_to_encoded_columns",
            "expected_getter",
        ),
        [
            # binary response
            (
                ["b", "d", "f"],
                None,
                {
                    "b": {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6},
                    "d": {1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6},
                    "f": {False: 2, True: 5},
                },
                {"b": ["b"], "d": ["d"], "f": ["f"]},
                expected_df_1,
            ),
            # multi level, single level
            (
                ["b", "f"],
                ["blue"],
                {
                    "b_blue": {"a": 1, "b": 1, "c": 0, "d": 0, "e": 0, "f": 0},
                    "f_blue": {False: 2 / 3, True: 0},
                },
                {"b": ["b_blue"], "f": ["f_blue"]},
                expected_df_2,
            ),
            # multi level, all levels
            (
                ["b", "f"],
                "all",
                {
                    "b_blue": {"a": 1, "b": 1, "c": 0, "d": 0, "e": 0, "f": 0},
                    "b_yellow": {"a": 0, "b": 0, "c": 1, "d": 1, "e": 0, "f": 0},
                    "b_green": {"a": 0, "b": 0, "c": 0, "d": 0, "e": 1, "f": 1},
                    "f_blue": {False: 2 / 3, True: 0},
                    "f_yellow": {False: 1 / 3, True: 1 / 3},
                    "f_green": {False: 0, True: 2 / 3},
                },
                {
                    "b": ["b_blue", "b_yellow", "b_green"],
                    "f": ["f_blue", "f_yellow", "f_green"],
                },
                expected_df_3,
            ),
        ],
    )
    def test_expected_outputs(
        self,
        library,
        columns,
        level,
        mappings,
        column_to_encoded_columns,
        expected_getter,
    ):
        """Test that the output is expected from transform with various parametrized setups."""

        df = create_MeanResponseTransformer_test_df(library=library)
        expected = expected_getter(library=library)

        x = MeanResponseTransformer(columns=columns, level=level)

        # set the impute values dict directly rather than fitting x on df so test works with helpers
        x.mappings = mappings

        x.column_to_encoded_columns = column_to_encoded_columns
        x.response_levels = level
        x.encoded_columns = list(x.mappings.keys())
        x.return_dtypes = {col: x.return_type for col in x.encoded_columns}

        df_transformed = x.transform(df)

        expected = nw.from_native(expected)
        for col in x.encoded_columns:
            expected = expected.with_columns(
                nw.col(col).cast(getattr(nw, x.return_type)),
            )
        expected = nw.to_native(expected)

        column_order = nw.from_native(df_transformed).columns

        assert_frame_equal_dispatch(
            df_transformed,
            expected[column_order],
        )

        # also test single row transform
        df = nw.from_native(df)
        expected = nw.from_native(expected)

        for i in range(len(df)):
            df_transformed_row = x.transform(df[[i]].to_native())
            df_expected_row = expected[[i]].to_native()

            column_order = nw.from_native(df_transformed_row).columns

            assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row[column_order],
            )

    # NOTE - this currently is more of a Fit test imo, but will leave in place for now
    # as does also test Transform
    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("columns", "target", "unseen_level_handling", "level", "expected_getter"),
        [
            (["b", "d", "f"], "a", "mean", None, expected_df_4),
            (["b", "d", "f"], "a", "median", None, expected_df_5),
            (["b", "d", "f"], "a", "min", None, expected_df_6),
            (["b", "d", "f"], "a", "max", None, expected_df_7),
            (["b", "d", "f"], "a", 21.6, None, expected_df_8),
            (["b", "f"], "multi_level_response", "mean", ["blue"], expected_df_9),
            (["b", "f"], "multi_level_response", "max", "all", expected_df_10),
        ],
    )
    def test_expected_outputs_with_unseen_level_handling(
        self,
        library,
        columns,
        target,
        unseen_level_handling,
        level,
        expected_getter,
    ):
        """Test that the output is expected from transform with various configs and unseen level handling"""

        df = create_MeanResponseTransformer_test_df_unseen_levels(library=library)
        expected = expected_getter(library=library)

        x = MeanResponseTransformer(
            columns=columns,
            level=level,
            unseen_level_handling=unseen_level_handling,
        )

        initial_df = create_MeanResponseTransformer_test_df()
        x.fit(initial_df, initial_df[target])

        df_transformed = x.transform(df)

        expected = nw.from_native(expected)
        for col in x.encoded_columns:
            expected = expected.with_columns(
                nw.col(col).cast(getattr(nw, x.return_type)),
            )

        column_order = nw.from_native(df_transformed).columns

        assert_frame_equal_dispatch(
            df_transformed,
            expected[column_order].to_native(),
        )

        # also test single row transform
        df = nw.from_native(df)
        expected = nw.from_native(expected)

        for i in range(len(df)):
            df_transformed_row = x.transform(df[[i]].to_native())
            df_expected_row = expected[[i]].to_native()

            column_order = nw.from_native(df_transformed_row).columns

            assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row[column_order],
            )

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_nulls_introduced_in_transform_error(self, library):
        """Test that transform will raise an error if nulls are introduced."""
        df = create_MeanResponseTransformer_test_df(library=library)

        x = MeanResponseTransformer(columns=["b", "d", "f"])

        x.fit(df, df["a"])

        df = nw.from_native(df)
        df = df.with_columns(nw.lit("z").alias("b"))
        df = nw.to_native(df)

        with pytest.raises(
            ValueError,
            match="MeanResponseTransformer: nulls would be introduced into column b from levels not present in mapping",
        ):
            x.transform(df)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        "prior, level, target, unseen_level_handling",
        [
            (5, "all", "c", "mean"),
            (100, ["a", "b"], "c", "min"),
            (1, None, "a", "max"),
            (0, None, "a", "median"),
        ],
    )
    def test_return_type_can_be_changed(
        self,
        prior,
        level,
        target,
        unseen_level_handling,
        library,
    ):
        "Test that output return types are controlled by return_type param, this defaults to float32 so test float64 here"

        df = create_MeanResponseTransformer_test_df(library=library)

        columns = ["b", "d", "f"]
        x = MeanResponseTransformer(
            columns=columns,
            return_type="Float64",
            prior=prior,
            unseen_level_handling=unseen_level_handling,
            level=level,
        )

        x.fit(df, df[target])

        output_df = nw.from_native(x.transform(df))

        if target == "c":
            actual_levels = df[target].unique().to_list() if level == "all" else level
            expected_created_cols = [
                prefix + "_" + suffix
                for prefix, suffix in product(columns, actual_levels)
            ]

        else:
            expected_created_cols = columns

        for col in expected_created_cols:
            expected_type = x.return_type
            actual_type = str(output_df[col].dtype)
            assert (
                actual_type == expected_type
            ), f"{x.classname} should output columns with type determine by the return_type param, expected {expected_type} but got {actual_type}"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_learnt_values_not_modified(self, library):
        """Test that the mappings from fit are not changed in transform."""
        df = create_MeanResponseTransformer_test_df(library=library)

        x = MeanResponseTransformer(columns="b")

        x.fit(df, df["a"])

        x2 = MeanResponseTransformer(columns="b")

        x2.fit(df, df["a"])

        x2.transform(df)

        assert (
            x.mappings == x2.mappings
        ), "Mean response values not changed in transform"


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MeanResponseTransformer"



from copy import deepcopy

import pandas as pd
import polars as pl
import pytest

from tests.utils import assert_frame_equal_dispatch


@pytest.mark.parametrize("library", ["pandas", "polars"])
def test_successful_outcome(library):
    "test outcome of function"

    dataframe_init = None
    if library == "pandas":
        dataframe_init = pd.DataFrame

    elif library == "polars":
        dataframe_init = pl.DataFrame

    df1 = dataframe_init({"a": [1, 2, 3], "b": [4, 5, 6]})

    df2 = deepcopy(df1)

    assert_frame_equal_dispatch(df1, df2)


@pytest.mark.parametrize("library", ["pandas", "polars"])
def test_failed_outcome(library):
    "test outcome of function"

    dataframe_init = None
    failed_assert_error = None
    if library == "pandas":
        dataframe_init = pd.DataFrame
        failed_assert_error = "DataFrame.columns values are different"

    elif library == "polars":
        dataframe_init = pl.DataFrame
        failed_assert_error = "DataFrames are different"

    df1 = dataframe_init({"a": [1, 2, 3], "b": [4, 5, 6]})

    df2 = dataframe_init({"b": [1, 2, 3], "a": [4, 5, 6]})

    with pytest.raises(
        AssertionError,
        match=failed_assert_error,
    ):
        assert_frame_equal_dispatch(df1, df2)



from collections import OrderedDict

import pandas as pd
import pytest
import test_aide as ta

import tests.test_data as d
from tests.base_tests import OtherBaseBehaviourTests
from tests.mapping.test_BaseCrossColumnMappingTransformer import (
    BaseCrossColumnMappingTransformerInitTests,
    BaseCrossColumnMappingTransformerTransformTests,
)
from tubular.mapping import CrossColumnMappingTransformer


class TestInit(BaseCrossColumnMappingTransformerInitTests):
    """Tests for CrossColumnMappingTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CrossColumnMappingTransformer"

    def test_mappings_not_ordered_dict_error(self):
        """Test that an exception is raised if mappings is not an ordered dict if more than 1 mapping is defined ."""
        with pytest.raises(
            TypeError,
            match="CrossColumnMappingTransformer: mappings should be an ordered dict for 'replace' mappings using multiple columns",
        ):
            CrossColumnMappingTransformer(
                mappings={"a": {"a": 1}, "b": {"b": 2}},
                adjust_column="c",
            )


class TestTransform(BaseCrossColumnMappingTransformerTransformTests):
    """Tests for the transform method on CrossColumnMappingTransformer."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CrossColumnMappingTransformer"

    def expected_df_1():
        """Expected output for test_expected_output."""
        return pd.DataFrame(
            {"a": [1, 2, 3, 4, 5, 6], "b": ["aa", "bb", "cc", "dd", "ee", "ff"]},
        )

    def expected_df_2():
        """Expected output for test_non_specified_values_unchanged."""
        return pd.DataFrame(
            {"a": [1, 2, 3, 4, 5, 6], "b": ["aa", "bb", "cc", "d", "e", "f"]},
        )

    def expected_df_3():
        """Expected output for test_multiple_mappings_ordered_dict."""
        return pd.DataFrame(
            {
                "a": [4, 2, 2, 1, 3],
                "b": ["x", "z", "y", "x", "x"],
                "c": ["cc", "dd", "bb", "cc", "cc"],
            },
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_1(), expected_df_1()),
    )
    def test_expected_output(self, df, expected):
        """Test that transform is giving the expected output."""
        mapping = {"a": {1: "aa", 2: "bb", 3: "cc", 4: "dd", 5: "ee", 6: "ff"}}

        x = CrossColumnMappingTransformer(mappings=mapping, adjust_column="b")

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="expected output from cross column mapping transformer",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_1(), expected_df_2()),
    )
    def test_non_specified_values_unchanged(self, df, expected):
        """Test that values not specified in mappings are left unchanged in transform."""
        mapping = {"a": {1: "aa", 2: "bb", 3: "cc"}}

        x = CrossColumnMappingTransformer(mappings=mapping, adjust_column="b")

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="expected output from cross column mapping transformer",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_7(), expected_df_3()),
    )
    def test_multiple_mappings_ordered_dict(self, df, expected):
        """Test that mappings by multiple columns using an ordered dict gives the expected output in transform."""
        mapping = OrderedDict()

        mapping["a"] = {1: "aa", 2: "bb"}
        mapping["b"] = {"x": "cc", "z": "dd"}

        x = CrossColumnMappingTransformer(mappings=mapping, adjust_column="c")

        df_transformed = x.transform(df)

        ta.equality.assert_frame_equal_msg(
            actual=df_transformed,
            expected=expected,
            msg_tag="expected output from cross column mapping transformer",
        )

    def test_mappings_unchanged(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that mappings is unchanged in transform - this overwrites a base test as
        logic specific to this transformer is needed."""
        df = d.create_df_1()

        mapping = {
            "a": {1: "a", 2: "b", 3: "c", 4: "d", 5: "e", 6: "f"},
            "b": {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6},
        }

        mapping = OrderedDict(mapping)

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["mappings"] = mapping

        x = uninitialized_transformers[self.transformer_name](**args)

        x.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=mapping,
            actual=x.mappings,
            msg=f"{self.transformer_name}.transform has changed self.mappings unexpectedly",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "CrossColumnMappingTransformer"



import joblib
import numpy as np
import pandas as pd
import pytest
import test_aide as ta

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
    DropOriginalTransformMixinTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
)
from tests.dates.test_BaseDatetimeTransformer import (
    DatetimeMixinTransformTests,
)
from tubular.dates import DatetimeInfoExtractor


@pytest.fixture()
def timeofday_extractor():
    return DatetimeInfoExtractor(columns=["a"], include=["timeofday"])


@pytest.fixture()
def timeofmonth_extractor():
    return DatetimeInfoExtractor(columns=["a"], include=["timeofmonth"])


@pytest.fixture()
def timeofyear_extractor():
    return DatetimeInfoExtractor(columns=["a"], include=["timeofyear"])


@pytest.fixture()
def dayofweek_extractor():
    return DatetimeInfoExtractor(columns=["a"], include=["dayofweek"])


class TestInit(
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
):
    "tests for DatetimeInfoExtractor.__init__"

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DatetimeInfoExtractor"

    @pytest.mark.parametrize("incorrect_type_include", [2, 3.0, "invalid", "dayofweek"])
    def test_error_when_include_not_list(self, incorrect_type_include):
        """Test that an exception is raised when value include variable is not a list."""
        with pytest.raises(
            TypeError,
            match="include should be List",
        ):
            DatetimeInfoExtractor(columns=["a"], include=incorrect_type_include)

    def test_error_when_invalid_include_option(self):
        """Test that an exception is raised when include contains incorrect values."""
        with pytest.raises(
            ValueError,
            match=r"DatetimeInfoExtractor: elements in include should be in \['timeofday', 'timeofmonth', 'timeofyear', 'dayofweek'\]",
        ):
            DatetimeInfoExtractor(
                columns=["a"],
                include=["timeofday", "timeofmonth", "invalid_option"],
            )

    @pytest.mark.parametrize(
        "incorrect_type_datetime_mappings",
        [2, 3.0, ["a", "b"], "dayofweek"],
    )
    def test_error_when_datetime_mappings_not_dict(
        self,
        incorrect_type_datetime_mappings,
    ):
        """Test that an exception is raised when datetime_mappings is not a dict."""
        with pytest.raises(
            TypeError,
            match="datetime_mappings should be Dict",
        ):
            DatetimeInfoExtractor(
                columns=["a"],
                datetime_mappings=incorrect_type_datetime_mappings,
            )

    @pytest.mark.parametrize(
        "incorrect_type_datetime_mappings_values",
        [{"timeofday": 2}],
    )
    def test_error_when_datetime_mapping_value_not_dict(
        self,
        incorrect_type_datetime_mappings_values,
    ):
        """Test that an exception is raised when values in datetime_mappings are not dict."""
        with pytest.raises(
            TypeError,
            match="values in datetime_mappings should be dict",
        ):
            DatetimeInfoExtractor(
                columns=["a"],
                datetime_mappings=incorrect_type_datetime_mappings_values,
            )

    @pytest.mark.parametrize(
        ("include", "incorrect_datetime_mappings_keys"),
        [
            (["timeofyear"], {"invalid_key": {"valid_mapping": "valid_output"}}),
            (["timeofyear"], {"dayofweek": {"day": range(7)}}),
            (
                ["timeofyear"],
                {"timeofyear": {"month": range(12)}, "timeofday": {"hour": range(24)}},
            ),
        ],
    )
    def test_error_when_datetime_mapping_key_not_in_include(
        self,
        include,
        incorrect_datetime_mappings_keys,
    ):
        """Test that an exception is raised when keys in datetime_mappings are not in include."""
        with pytest.raises(
            ValueError,
            match="keys in datetime_mappings should be in include",
        ):
            DatetimeInfoExtractor(
                columns=["a"],
                include=include,
                datetime_mappings=incorrect_datetime_mappings_keys,
            )

    @pytest.mark.parametrize(
        ("incomplete_mappings", "expected_exception"),
        [
            (
                {"timeofday": {"mapped": range(23)}},
                r"DatetimeInfoExtractor: timeofday mapping dictionary should contain mapping for all values between 0-23. \{23\} are missing",
            ),
            (
                {"timeofmonth": {"mapped": range(1, 31)}},
                r"DatetimeInfoExtractor: timeofmonth mapping dictionary should contain mapping for all values between 1-31. \{31\} are missing",
            ),
            (
                {"timeofyear": {"mapped": range(1, 12)}},
                r"DatetimeInfoExtractor: timeofyear mapping dictionary should contain mapping for all values between 1-12. \{12\} are missing",
            ),
            (
                {"dayofweek": {"mapped": range(6)}},
                r"DatetimeInfoExtractor: dayofweek mapping dictionary should contain mapping for all values between 0-6. \{6\} are missing",
            ),
        ],
    )
    def test_error_when_incomplete_mappings_passed(
        self,
        incomplete_mappings,
        expected_exception,
    ):
        """Test that error is raised when incomplete mappings are passed."""
        with pytest.raises(ValueError, match=expected_exception):
            DatetimeInfoExtractor(columns=["a"], datetime_mappings=incomplete_mappings)


class TestMapValues:
    "tests for DatetimeInfoExtractor.map_values"

    @pytest.mark.parametrize("incorrect_type_input", ["2", [1, 2]])
    def test_incorrect_type_input(self, incorrect_type_input, timeofday_extractor):
        """Test that an error is raised if input is the wrong type."""
        with pytest.raises(
            ValueError,
            match="DatetimeInfoExtractor: value for timeofday mapping in self._map_values should be an integer value in 0-23",
        ):
            timeofday_extractor._map_values(incorrect_type_input, "timeofday")

    @pytest.mark.parametrize("incorrect_size_input", [-2, 30, 5.6, 11.2])
    def test_out_of_bounds_or_fractional_input(
        self,
        incorrect_size_input,
        timeofday_extractor,
    ):
        """Test that an error is raised when value is outside of 0-23 range."""
        with pytest.raises(
            ValueError,
            match="DatetimeInfoExtractor: value for timeofday mapping in self._map_values should be an integer value in 0-23",
        ):
            timeofday_extractor._map_values(incorrect_size_input, "timeofday")

    @pytest.mark.parametrize(
        ("valid_hour", "hour_time_of_day"),
        [
            (0, "night"),
            (5, "night"),
            (6, "morning"),
            (11, "morning"),
            (12, "afternoon"),
            (17, "afternoon"),
            (18, "evening"),
            (23, "evening"),
        ],
    )
    def test_valid_inputs_timeofday(
        self,
        valid_hour,
        hour_time_of_day,
        timeofday_extractor,
    ):
        """Trial test to check all in one go."""
        output = timeofday_extractor._map_values(valid_hour, "timeofday")

        assert output == hour_time_of_day, "expected {}, output {}".format(
            hour_time_of_day,
            output,
        )

    @pytest.mark.parametrize(
        ("valid_day", "day_time_of_month"),
        [
            (1, "start"),
            (6, "start"),
            (10, "start"),
            (11, "middle"),
            (16, "middle"),
            (20, "middle"),
            (21, "end"),
            (21, "end"),
            (31, "end"),
        ],
    )
    def test_valid_inputs_timeofmonth(
        self,
        valid_day,
        day_time_of_month,
        timeofmonth_extractor,
    ):
        """Test that correct values are return with valid inputs."""
        output = timeofmonth_extractor._map_values(valid_day, "timeofmonth")
        assert output == day_time_of_month, "expected {}, output {}".format(
            day_time_of_month,
            output,
        )

    @pytest.mark.parametrize(
        ("valid_month", "month_time_of_year"),
        [
            (1, "winter"),
            (3, "spring"),
            (4, "spring"),
            (6, "summer"),
            (7, "summer"),
            (9, "autumn"),
            (10, "autumn"),
            (12, "winter"),
        ],
    )
    def test_valid_inputs_timeofyear(
        self,
        valid_month,
        month_time_of_year,
        timeofyear_extractor,
    ):
        """Test that correct values are return with valid inputs."""
        output = timeofyear_extractor._map_values(valid_month, "timeofyear")
        assert output == month_time_of_year, "expected {}, output {}".format(
            month_time_of_year,
            output,
        )

    @pytest.mark.parametrize(
        ("valid_day", "dayofweek"),
        [
            (0, "monday"),
            (2, "wednesday"),
            (4, "friday"),
            (6, "sunday"),
        ],
    )
    def test_valid_inputs_dayofweek(self, valid_day, dayofweek, dayofweek_extractor):
        """Test that correct values are return with valid inputs."""
        output = dayofweek_extractor._map_values(valid_day, "dayofweek")
        assert output == dayofweek, f"expected {dayofweek}, output {output}"

    def test_valid_nan_output(self, timeofday_extractor):
        """Test that correct values are return with valid inputs."""
        output = timeofday_extractor._map_values(np.nan, "timeofday")

        assert np.isnan(
            output,
        ), f"passing np.nan should result in np.nan, instead received {output}"


class TestTransform(
    GenericTransformTests,
    DatetimeMixinTransformTests,
    DropOriginalTransformMixinTests,
):
    "tests for DatetimeInfoExtractor.transform"

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DatetimeInfoExtractor"

    def test_correct_col_returned(self):
        """Test that the added column is correct."""
        df = d.create_date_test_df()
        df = df.astype("datetime64[ns]")

        x = DatetimeInfoExtractor(columns=["b"], include=["timeofyear"])
        transformed = x.transform(df)

        expected_output = pd.Series(
            [
                "spring",
                "winter",
                "autumn",
                "autumn",
                "autumn",
                "autumn",
                "autumn",
                "summer",
            ],
            name="b_timeofyear",
        )

        ta.equality.assert_series_equal_msg(
            transformed["b_timeofyear"],
            expected_output,
            "incorrect series returned",
            print_actual_and_expected=True,
        )

    def test_map_values_calls(self, mocker):
        """Test all intermediary methods are being called correct number of times."""
        # df is 8 rows long so each intermediate function must have 8 calls
        df = d.create_date_test_df()
        df = df.astype("datetime64[ns]")

        mocked_map_values = mocker.spy(DatetimeInfoExtractor, "_map_values")

        x = DatetimeInfoExtractor(
            columns=["b"],
            include=["timeofday", "timeofyear", "timeofmonth", "dayofweek"],
        )
        x.transform(df)

        assert mocked_map_values.call_count == 32

    def test_correct_df_returned_datetime_input(self):
        """Test that correct df is returned after transformation."""
        df = d.create_date_test_df()
        df.loc[0, "b"] = np.nan
        df = df.astype("datetime64[ns]")

        x = DatetimeInfoExtractor(columns=["b"], include=["timeofmonth", "timeofyear"])
        transformed = x.transform(df)

        expected = df.copy()
        expected["b_timeofmonth"] = [
            np.nan,
            "end",
            "start",
            "start",
            "start",
            "start",
            "start",
            "end",
        ]
        expected["b_timeofyear"] = [
            np.nan,
            "winter",
            "autumn",
            "autumn",
            "autumn",
            "autumn",
            "autumn",
            "summer",
        ]

        ta.equality.assert_frame_equal_msg(
            transformed,
            expected,
            "incorrect dataframe returned",
        )

    def test_is_serialisable(self, tmp_path):
        transformer = DatetimeInfoExtractor(columns=["b"], include=["timeofyear"])

        # pickle transformer
        path = tmp_path / "transformer.pkl"

        # serialise without raising error
        joblib.dump(transformer, path)


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "DatetimeInfoExtractor"



import narwhals as nw
import pytest

import tests.test_data as d
from tests import utils as u
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
)
from tubular.imputers import NullIndicator


class TestInit(ColumnStrListInitTests):
    """Tests for NullIndicator.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "NullIndicator"


class TestTransform(GenericTransformTests):
    """Tests for NullIndicator.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "NullIndicator"

    @pytest.fixture()
    def expected_df_1(self, request):
        """Expected output for test_null_indicator_columns_correct."""
        library = request.param

        df_dict1 = {
            "a": [1, 2, None, 4, None, 6],
            "b": [None, 5, 4, 3, 2, 1],
            "c": [3, 2, 1, 4, 5, 6],
            "b_nulls": [1, 0, 0, 0, 0, 0],
            "c_nulls": [0, 0, 0, 0, 0, 0],
        }

        df1 = u.dataframe_init_dispatch(dataframe_dict=df_dict1, library=library)

        narwhals_df = nw.from_native(df1)

        # Convert adjusted expected columns to Boolean
        for col in ["b_nulls", "c_nulls"]:
            narwhals_df = narwhals_df.with_columns(
                narwhals_df[col].cast(nw.Boolean),
            )

        return narwhals_df.to_native()

    @pytest.mark.parametrize(
        ("library", "expected_df_1"),
        [("pandas", "pandas"), ("polars", "polars")],
        indirect=["expected_df_1"],
    )
    def test_null_indicator_columns_correct(self, expected_df_1, library):
        """Test that the created indicator column is correct - and unrelated columns are unchanged."""
        df = d.create_df_9(library=library)

        columns = ["b", "c"]
        transformer = NullIndicator(columns=columns)

        df_transformed = transformer.transform(df)

        # Convert both DataFrames to a common format using Narwhals
        df_transformed_common = nw.from_native(df_transformed)
        expected_df_1_common = nw.from_native(expected_df_1)

        # Check outcomes for single rows
        for i in range(len(df_transformed_common)):
            df_transformed_row = df_transformed_common[[i]].to_native()
            df_expected_row = expected_df_1_common[[i]].to_native()

            u.assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row,
            )

        # Check whole dataframes
        u.assert_frame_equal_dispatch(
            df_transformed_common.to_native(),
            expected_df_1_common.to_native(),
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "NullIndicator"



import narwhals as nw
import numpy as np
import pytest

from tests.test_data import create_df_2
from tubular.mixins import WeightColumnMixin


class TestCheckAndSetWeight:
    @pytest.mark.parametrize("weights_column", (0, ["a"], {"a": 10}))
    def test_weight_arg_errors(
        self,
        weights_column,
    ):
        """Test that appropriate errors are throw for bad weight arg."""

        obj = WeightColumnMixin()

        with pytest.raises(
            TypeError,
            match="weights_column should be str or None",
        ):
            obj.check_and_set_weight(weights_column)


class TestCheckWeightsColumn:
    @pytest.mark.parametrize(
        "library",
        [
            "pandas",
            "polars",
        ],
    )
    @pytest.mark.parametrize(
        "bad_weight_value, expected_message",
        [
            (None, "weight column must be non-null"),
            (np.inf, "weight column must not contain infinite values."),
            (-np.inf, "weight column must be positive"),
            (-1, "weight column must be positive"),
        ],
    )
    def test_bad_values_in_weights_error(
        self,
        bad_weight_value,
        expected_message,
        library,
    ):
        """Test that an exception is raised if there are negative/nan/inf values in sample_weight."""

        df = create_df_2(library=library)

        obj = WeightColumnMixin()

        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df)

        weight_column = "weight_column"

        df = df.with_columns(
            nw.new_series(
                weight_column,
                [*[bad_weight_value], *np.arange(2, len(df) + 1)],
                backend=native_backend,
            ),
        )

        df = nw.to_native(df)

        with pytest.raises(ValueError, match=expected_message):
            obj.check_weights_column(df, weight_column)

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_weight_col_non_numeric(
        self,
        library,
    ):
        """Test an error is raised if weight is not numeric."""

        obj = WeightColumnMixin()

        df = create_df_2(library=library)
        df = nw.from_native(df)

        weight_column = "weight_column"
        error = r"weight column must be numeric."
        df = df.with_columns(nw.lit("a").alias(weight_column))
        df = nw.to_native(df)

        with pytest.raises(
            ValueError,
            match=error,
        ):
            # using check_weights_column method to test correct error is raised for transformers that use weights

            obj.check_weights_column(df, weight_column)

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_weight_not_in_X_error(
        self,
        library,
    ):
        """Test an error is raised if weight is not in X"""

        obj = WeightColumnMixin()

        df = create_df_2(library=library)

        weight_column = "weight_column"
        error = rf"weight col \({weight_column}\) is not present in columns of data"

        with pytest.raises(
            ValueError,
            match=error,
        ):
            # using check_weights_column method to test correct error is raised for transformers that use weights

            obj.check_weights_column(df, weight_column)

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    def test_zero_total_weight_error(
        self,
        library,
    ):
        """Test that an exception is raised if the total sample weights are 0."""

        obj = WeightColumnMixin()

        weight_column = "weight_column"

        df = create_df_2(library=library)

        df = nw.from_native(df)
        df = df.with_columns(nw.lit(0).alias(weight_column))
        df = nw.to_native(df)

        with pytest.raises(
            ValueError,
            match="total sample weights are not greater than 0",
        ):
            obj.check_weights_column(df, weight_column)



import re

import pytest

from tests.test_data import (
    create_bool_and_float_df,
    create_df_2,
    create_df_with_none_and_nan_cols,
    create_is_between_dates_df_1,
)
from tubular.mixins import CheckNumericMixin


class TestCheckNumericMixin:
    "tests for CheckNumericMixin class"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("df_generator", "bad_cols"),
        [
            (create_df_2, ["b", "c"]),  # str
            (create_is_between_dates_df_1, ["a"]),  # datetime
            (create_bool_and_float_df, ["b"]),  # bool
            (create_df_with_none_and_nan_cols, ["b"]),  # None
        ],
    )
    def test_check_numeric_columns_errors(self, library, df_generator, bad_cols):
        "test check_numeric_columns method raises appropriate error"

        df = df_generator(library=library)

        obj = CheckNumericMixin()

        # this object is generally wrapped in a transformer with a .columns attr, set this here
        obj.columns = bad_cols

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"CheckNumericMixin: The following columns are not numeric in X; {bad_cols}",
            ),
        ):
            obj.check_numeric_columns(df)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("df_generator", "cols"),
        [
            (create_df_2, ["a"]),  # int
            (create_bool_and_float_df, ["a"]),  # float
            (create_df_with_none_and_nan_cols, ["a"]),  # nan
        ],
    )
    def test_check_numeric_columns_passes(self, library, df_generator, cols):
        "test check_numeric_columns method passes for numeric columns"

        df = df_generator(library=library)

        obj = CheckNumericMixin()

        # this object is generally wrapped in a transformer with a .columns attr, set this here
        obj.columns = cols

        obj.check_numeric_columns(df)



import copy

import polars as pl
import pytest
from sklearn.exceptions import NotFittedError

import tests.test_data as d
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
)
from tests.utils import assert_frame_equal_dispatch


# The first part of this file builds out the tests for BaseNominalTransformer so that they can be
# imported into other test files (by not starting the class name with Test)
# The second part actually calls these tests (along with all other require tests) for the BaseNominalTransformer
class GenericNominalTransformTests(GenericTransformTests):
    """
    Tests for BaseNominalTransformer.transform().
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_not_fitted_error_raised(self, initialized_transformers, library):
        if initialized_transformers[self.transformer_name].FITS:
            df = d.create_df_1(library=library)

            transformer = initialized_transformers[self.transformer_name]

            # skip polars test if transformer not yet converted for polars
            if not transformer.polars_compatible and isinstance(df, pl.DataFrame):
                return

            with pytest.raises(NotFittedError):
                initialized_transformers[self.transformer_name].transform(df)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_non_mappable_rows_exception_raised(
        self,
        initialized_transformers,
        library,
    ):
        """Test an exception is raised if non-mappable rows are present in X."""
        df = d.create_df_1(library=library)

        transformer = initialized_transformers[self.transformer_name]

        # skip polars test if transformer not yet converted for polars
        if not transformer.polars_compatible and isinstance(df, pl.DataFrame):
            return

        transformer.fit(df)

        transformer.mappings = {
            "a": {1: 2, 2: 3, 3: 4, 4: 5, 5: 6, 6: 7},
            "b": {"a": 1, "c": 2, "d": 3, "e": 4, "f": 5},
        }

        with pytest.raises(
            ValueError,
            match=f"{self.transformer_name}: nulls would be introduced into column b from levels not present in mapping",
        ):
            transformer.transform(df)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_original_df_not_updated(self, initialized_transformers, library):
        """Test that the original dataframe is not transformed when transform method used."""

        df = d.create_df_1(library=library)

        transformer = initialized_transformers[self.transformer_name]

        # skip polars test if transformer not yet converted for polars
        if not transformer.polars_compatible and isinstance(df, pl.DataFrame):
            return

        transformer = transformer.fit(df)

        transformer.mappings = {"b": {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6}}

        _ = transformer.transform(df)

        assert_frame_equal_dispatch(df, d.create_df_1(library=library))

    @pytest.mark.parametrize(
        "library",
        ["pandas"],
    )
    def test_pandas_index_not_updated(
        self,
        initialized_transformers,
        library,
    ):
        """Test that the original (pandas) dataframe index is not transformed when transform method used."""

        df = d.create_df_1(library=library)

        x = initialized_transformers[self.transformer_name]

        x = x.fit(df)

        x.mappings = {"b": {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6}}

        # update to abnormal index
        df.index = [2 * i for i in df.index]

        original_df = copy.deepcopy(df)

        _ = x.transform(df)

        assert_frame_equal_dispatch(df, original_df)


class TestInit(ColumnStrListInitTests):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseNominalTransformer"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseNominalTransformer"


class TestTransform(GenericNominalTransformTests):
    """Tests for BaseImputer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseNominalTransformer"


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseNominalTransformer"



import narwhals as nw
import pytest

import tests.test_data as d
from tests.base_tests import OtherBaseBehaviourTests
from tests.capping.test_BaseCappingTransformer import (
    GenericCappingFitTests,
    GenericCappingInitTests,
    GenericCappingTransformTests,
)
from tests.utils import assert_frame_equal_dispatch, dataframe_init_dispatch
from tubular.capping import OutOfRangeNullTransformer


class TestInit(GenericCappingInitTests):
    """Tests for OutOfRangeNullTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OutOfRangeNullTransformer"


class TestFit(GenericCappingFitTests):
    """Tests for OutOfRangeNullTransformer.fit()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OutOfRangeNullTransformer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    @pytest.mark.parametrize(
        ("values", "sample_weight", "quantiles"),
        [
            ([1, 2, 3], [1, 2, 1], [0.1, 0.5]),
            ([1, 2, 3], None, [0.1, 0.5]),
            ([2, 4, 6, 8], [3, 2, 1, 1], [None, 0.5]),
            ([2, 4, 6, 8], None, [None, 0.5]),
            ([-1, -5, -10, 20, 30], [1, 2, 2, 2, 2], [0.1, None]),
            ([-1, -5, -10, 20, 40], None, [0.1, None]),
        ],
    )
    def test_replacement_values_updated(
        self,
        values,
        sample_weight,
        quantiles,
        minimal_attribute_dict,
        uninitialized_transformers,
        library,
    ):
        """Test that weighted_quantile gives the expected outputs."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["quantiles"] = {"a": quantiles}
        args["capping_values"] = None
        args["weights_column"] = "w"

        transformer = uninitialized_transformers[self.transformer_name](**args)

        if not sample_weight:
            sample_weight = [1] * len(values)

        df_dict = {
            "a": values,
            "w": sample_weight,
        }

        df = dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

        transformer.fit(df)

        lower_replacement = None if quantiles[0] else False
        upper_replacement = None if quantiles[1] else False
        expected = [lower_replacement, upper_replacement]
        assert (
            transformer._replacement_values["a"] == expected
        ), f"unexpected value for replacement_values attribute, expected {expected} but got {transformer._replacement_values}"


class TestTransform(GenericCappingTransformTests):
    """Tests for OutOfRangeNullTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OutOfRangeNullTransformer"

    def expected_df_1(self, library="pandas"):
        """Expected output from test_expected_output_min_and_max."""

        df_dict = {
            "a": [None, 2, 3, 4, 5, None, None],
            "b": [1, 2, 3, None, 7, None, None],
            "c": [None, 1, 2, 3, None, None, None],
        }

        return dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_expected_output_min_and_max_combinations(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
        library,
    ):
        """Test that capping is applied correctly in transform."""

        df = d.create_df_3(library=library)
        expected = self.expected_df_1(library=library)

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["capping_values"] = {"a": [2, 5], "b": [None, 7], "c": [0, None]}

        transformer = uninitialized_transformers[self.transformer_name](**args)

        df_transformed = transformer.transform(df)

        assert_frame_equal_dispatch(df_transformed, expected)

        # Check outcomes for single rows
        df = nw.from_native(df)
        expected = nw.from_native(expected)
        for i in range(len(df)):
            df_transformed_row = transformer.transform(df[[i]].to_native())
            df_expected_row = expected[[i]].to_native()

            assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row,
            )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "OutOfRangeNullTransformer"


class TestSetReplacementValues:
    """Test for the OutOfRangeNullTransformer.set_replacement_values() method."""

    @pytest.mark.parametrize(
        ("capping_values", "expected_replacement_values"),
        [
            (
                {"a": [0, 1], "b": [None, 1], "c": [3, None]},
                {"a": [None, None], "b": [False, None], "c": [None, False]},
            ),
            ({}, {}),
            ({"a": [None, 0.1]}, {"a": [False, None]}),
        ],
    )
    def test_expected_replacement_values_output(
        self,
        capping_values,
        expected_replacement_values,
    ):
        """Test the _replacement_values attribute is modified as expected given the prior values of the attribute."""

        replacement_values = OutOfRangeNullTransformer.set_replacement_values(
            capping_values,
        )

        assert (
            replacement_values == expected_replacement_values
        ), f"set_replacement_values output not as expected, expected {expected_replacement_values} but got {replacement_values}"



# tests to apply to all columns str or list transformers
import copy
import re

import joblib
import narwhals as nw
import numpy as np
import pandas as pd
import polars as pl
import pytest
import sklearn.base as b
from beartype.roar import BeartypeCallHintParamViolation

from tests.utils import assert_frame_equal_dispatch


class GenericInitTests:
    """
    Generic tests for transformer.init(). This test class does not contain tests for the behaviours
    associated with the "columns" argument because the structure of this argument varies between
    transformers. In this file are other test classes that inherit from this one which are specific
    to the different "columns" argument structures. Please choose the appropriate one of them to inherit
    when writing tests unless the transformer is a special case and needs unique tests written for
    "columns", in which case inherit this class.

    Note this class name deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    def test_print(self, initialized_transformers):
        """Test that transformer can be printed.
        If an error is raised in this test it will not prevent the transformer from working correctly,
        but will stop other unit tests passing.
        """

        print(initialized_transformers[self.transformer_name])

    def test_clone(self, initialized_transformers):
        """Test that transformer can be used in sklearn.base.clone function."""

        b.clone(initialized_transformers[self.transformer_name])

    @pytest.mark.parametrize("non_bool", [1, "True", {"a": 1}, [1, 2], None])
    def test_verbose_non_bool_error(
        self,
        non_bool,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test an error is raised if verbose is not specified as a bool."""

        with pytest.raises(
            BeartypeCallHintParamViolation,
        ):
            uninitialized_transformers[self.transformer_name](
                verbose=non_bool,
                **minimal_attribute_dict[self.transformer_name],
            )


class ColumnStrListInitTests(GenericInitTests):
    """
    Tests for BaseTransformer.init() behaviour specific to when a transformer takes columns as string or list.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    def test_columns_empty_list_error(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test an error is raised if columns is specified as an empty list."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["columns"] = []

        with pytest.raises(ValueError):
            uninitialized_transformers[self.transformer_name](**args)

    @pytest.mark.parametrize(
        "non_string",
        [1, True, {"a": 1}, [1, 2], None, np.inf, np.nan],
    )
    def test_columns_list_element_error(
        self,
        non_string,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test an error is raised if columns list contains non-string elements."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["columns"] = [non_string, non_string]

        with pytest.raises(
            BeartypeCallHintParamViolation,
        ):
            uninitialized_transformers[self.transformer_name](**args)

    @pytest.mark.parametrize(
        "non_string_or_list",
        [
            1,
            True,
            {"a": 1},
            None,
            np.inf,
            np.nan,
        ],
    )
    def test_columns_non_string_or_list_error(
        self,
        non_string_or_list,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test an error is raised if columns is not passed as a string or list."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["columns"] = non_string_or_list

        with pytest.raises(
            BeartypeCallHintParamViolation,
        ):
            uninitialized_transformers[self.transformer_name](**args)


class DropOriginalInitMixinTests:
    """
    Tests for BaseTransformer.init() behaviour specific to when a transformer accepts a "drop_original" column.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize("drop_orginal_column", (0, "a", ["a"], {"a": 10}, None))
    def test_drop_column_arg_errors(
        self,
        uninitialized_transformers,
        minimal_attribute_dict,
        drop_orginal_column,
    ):
        """Test that appropriate errors are throwm for non boolean arg."""
        args = minimal_attribute_dict[self.transformer_name].copy()
        args["drop_original"] = drop_orginal_column

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: drop_original should be bool",
        ):
            uninitialized_transformers[self.transformer_name](**args)


class NewColumnNameInitMixintests:
    """
    Tests for BaseTransformer.init() behaviour specific to when a transformer accepts a "new_column_name" column.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize(
        "new_column_type",
        [1, True, {"a": 1}, [1, 2], None, np.inf, np.nan],
    )
    def test_new_column_name_type_error(
        self,
        new_column_type,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test an error is raised if any type other than str passed to new_column_name"""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["new_column_name"] = new_column_type

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"{self.transformer_name}: new_column_name should be str",
            ),
        ):
            uninitialized_transformers[self.transformer_name](**args)


class SeparatorInitMixintests:
    """
    Tests for BaseTransformer.init() behaviour specific to when a transformer accepts a "separator" column.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize(
        "separator",
        [1, True, {"a": 1}, [1, 2], None, np.inf, np.nan],
    )
    def test_separator_type_error(
        self,
        separator,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test an error is raised if any type other than str passed to separator"""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["separator"] = separator

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"{self.transformer_name}: separator should be str",
            ),
        ):
            uninitialized_transformers[self.transformer_name](**args)


class WeightColumnInitMixinTests:
    """
    Tests for BaseTransformer.init() behaviour specific to when a transformer takes accepts a weight column.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize("weights_column", (0, ["a"], {"a": 10}))
    def test_weight_arg_errors(
        self,
        uninitialized_transformers,
        minimal_attribute_dict,
        weights_column,
    ):
        """Test that appropriate errors are throw for bad weight arg."""
        args = minimal_attribute_dict[self.transformer_name].copy()
        args["weights_column"] = weights_column

        with pytest.raises(
            TypeError,
            match="weights_column should be str or None",
        ):
            uninitialized_transformers[self.transformer_name](**args)


class TwoColumnListInitTests(ColumnStrListInitTests):
    """
    Tests for BaseTransformer.init() behaviour specific to when a transformer takes two columns as a list.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize("non_list", ["a", "b", "c"])
    def test_columns_non_list_error(
        self,
        non_list,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test an error is raised if columns is passed as a string not a list."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["columns"] = non_list

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"{self.transformer_name}: columns should be list",
            ),
        ):
            uninitialized_transformers[self.transformer_name](**args)

    @pytest.mark.parametrize("list_length", [["a", "a", "a"], ["a"]])
    def test_list_length_error(
        self,
        list_length,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test an error is raised if list of any length other than 2 is passed"""

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["columns"] = list_length

        with pytest.raises(
            ValueError,
            match=re.escape(
                f"{self.transformer_name}: This transformer works with two columns only",
            ),
        ):
            uninitialized_transformers[self.transformer_name](**args)


class GenericFitTests:
    """
    Generic tests for transformer.fit().
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_fit_returns_self(
        self,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test fit returns self?."""

        df = minimal_dataframe_lookup[self.transformer_name]

        x = initialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        x_fitted = x.fit(df, df["a"])

        assert (
            x_fitted is x
        ), f"Returned value from {self.transformer_name}.fit not as expected."

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_fit_not_changing_data(
        self,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test fit does not change X."""

        df = minimal_dataframe_lookup[self.transformer_name]
        x = initialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        original_df = copy.deepcopy(df)

        x.fit(df, df["a"])

        assert_frame_equal_dispatch(
            original_df,
            df,
        )

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    @pytest.mark.parametrize("non_df", [1, True, "a", [1, 2], {"a": 1}, None])
    def test_X_non_df_error(
        self,
        initialized_transformers,
        non_df,
        minimal_dataframe_lookup,
    ):
        """Test an error is raised if X is not passed as a pd/pl.DataFrame."""

        df = minimal_dataframe_lookup[self.transformer_name]
        x = initialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: X should be a polars or pandas DataFrame/LazyFrame",
        ):
            x.fit(non_df, df["a"])

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    @pytest.mark.parametrize("non_series", [1, True, "a", [1, 2], {"a": 1}])
    def test_bad_type_error(
        self,
        non_series,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test an error is raised if y is not passed as a pd/pl.Series."""

        df = minimal_dataframe_lookup[self.transformer_name]
        x = initialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: unexpected type for y, should be a polars or pandas Series",
        ):
            x.fit(X=df, y=non_series)

    def test_X_no_rows_error(
        self,
        initialized_transformers,
    ):
        """Test an error is raised if X has no rows."""

        x = initialized_transformers[self.transformer_name]

        df = pd.DataFrame(columns=["a", "b", "c"])

        with pytest.raises(
            ValueError,
            match=re.escape(f"{self.transformer_name}: X has no rows; (0, 3)"),
        ):
            x.fit(df, df["a"])

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_Y_no_rows_error(
        self,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test an error is raised if Y has no rows."""

        x = initialized_transformers[self.transformer_name]

        df = minimal_dataframe_lookup[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        if isinstance(df, pd.DataFrame):
            series_init = pd.Series
        elif isinstance(df, pl.DataFrame):
            series_init = pl.Series
        else:
            series_init = None

        with pytest.raises(
            ValueError,
            match=re.escape(f"{self.transformer_name}: y is empty; (0,)"),
        ):
            x.fit(X=df, y=series_init(name="d", dtype=object))

    def test_unexpected_kwarg_error(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "__init__() got an unexpected keyword argument 'unexpected_kwarg'",
            ),
        ):
            uninitialized_transformers[self.transformer_name](
                unexpected_kwarg=True,
                **minimal_attribute_dict[self.transformer_name],
            )


class CheckNumericFitMixinTests:
    """
    Tests for BaseTransformer.init() behaviour specific to when a transformer used.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    def test_exception_raised(self, initialized_transformers, minimal_dataframe_lookup):
        """Test an exception is raised if non numeric columns are passed in X."""
        df = minimal_dataframe_lookup[self.transformer_name]
        df["a"] = "string"

        x = initialized_transformers[self.transformer_name]

        with pytest.raises(
            TypeError,
            match=re.escape(
                f"{self.transformer_name}: The following columns are not numeric in X; ['a']",
            ),
        ):
            x.fit(df)


class WeightColumnFitMixinTests:
    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_fit_returns_self_weighted(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test fit returns self?."""
        df = minimal_dataframe_lookup[self.transformer_name]

        uninitialized_transformer = uninitialized_transformers[self.transformer_name]
        # skip polars test if not narwhalified
        if not uninitialized_transformer.polars_compatible and isinstance(
            df,
            pl.DataFrame,
        ):
            return

        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df)

        args = minimal_attribute_dict[self.transformer_name].copy()
        # insert weight column
        weight_column = "weight_column"
        args["weights_column"] = weight_column
        df = df.with_columns(
            nw.new_series(
                weight_column,
                np.arange(1, len(df) + 1),
                backend=native_backend,
            ),
        )

        df = nw.to_native(df)

        transformer = uninitialized_transformer(**args)

        x_fitted = transformer.fit(df, df["a"])

        assert (
            x_fitted is transformer
        ), f"Returned value from {self.transformer_name}.fit not as expected."

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_fit_not_changing_data_weighted(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test fit does not change X - when weights are used."""
        df = minimal_dataframe_lookup[self.transformer_name]
        uninitialized_transformer = uninitialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not uninitialized_transformer.polars_compatible and isinstance(
            df,
            pl.DataFrame,
        ):
            return

        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df)
        # insert weight column
        weight_column = "weight_column"
        df = df.with_columns(
            nw.new_series(
                weight_column,
                np.arange(1, len(df) + 1),
                backend=native_backend,
            ),
        )

        original_df = df.clone()

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["weights_column"] = weight_column

        transformer = uninitialized_transformer(**args)

        df = nw.to_native(df)
        original_df = nw.to_native(original_df)

        transformer.fit(df, df["a"])

        assert_frame_equal_dispatch(df, original_df)

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "bad_weight_value, expected_message",
        [
            (np.nan, "weight column must be non-null"),
            (None, "weight column must be non-null"),
            (np.inf, "weight column must not contain infinite values."),
            (-np.inf, "weight column must be positive"),
            (-1, "weight column must be positive"),
        ],
    )
    def test_bad_values_in_weights_error(
        self,
        bad_weight_value,
        expected_message,
        minimal_attribute_dict,
        uninitialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test that an exception is raised if there are negative/nan/inf values in sample_weight."""

        df = minimal_dataframe_lookup[self.transformer_name]
        uninitialized_transformer = uninitialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not uninitialized_transformer.polars_compatible and isinstance(
            df,
            pl.DataFrame,
        ):
            return

        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df)

        args = minimal_attribute_dict[self.transformer_name].copy()
        weight_column = "weight_column"
        args["weights_column"] = weight_column

        df = df.with_columns(
            nw.new_series(
                weight_column,
                [*[bad_weight_value], *np.arange(2, len(df) + 1)],
                backend=native_backend,
            ),
        )

        df = nw.to_native(df)

        transformer = uninitialized_transformer(**args)

        with pytest.raises(ValueError, match=expected_message):
            transformer.fit(df, df["a"])

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_weight_col_non_numeric(
        self,
        uninitialized_transformers,
        minimal_attribute_dict,
        minimal_dataframe_lookup,
    ):
        """Test an error is raised if weight is not numeric."""

        df = minimal_dataframe_lookup[self.transformer_name]
        uninitialized_transformer = uninitialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not uninitialized_transformer.polars_compatible and isinstance(
            df,
            pl.DataFrame,
        ):
            return

        df = nw.from_native(df)

        weight_column = "weight_column"
        error = r"weight column must be numeric."
        df = df.with_columns(nw.lit("a").alias(weight_column))
        df = nw.to_native(df)

        with pytest.raises(
            ValueError,
            match=error,
        ):
            # using check_weights_column method to test correct error is raised for transformers that use weights

            args = minimal_attribute_dict[self.transformer_name].copy()
            args["weights_column"] = weight_column

            transformer = uninitialized_transformer(**args)
            transformer.fit(df, df["a"])

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_weight_not_in_X_error(
        self,
        uninitialized_transformers,
        minimal_attribute_dict,
        minimal_dataframe_lookup,
    ):
        """Test an error is raised if weight is not in X"""

        df = minimal_dataframe_lookup[self.transformer_name]
        uninitialized_transformer = uninitialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not uninitialized_transformer.polars_compatible and isinstance(
            df,
            pl.DataFrame,
        ):
            return

        weight_column = "weight_column"
        error = rf"weight col \({weight_column}\) is not present in columns of data"

        with pytest.raises(
            ValueError,
            match=error,
        ):
            # using check_weights_column method to test correct error is raised for transformers that use weights

            args = minimal_attribute_dict[self.transformer_name].copy()
            args["weights_column"] = weight_column

            transformer = uninitialized_transformer(**args)
            transformer.fit(df, df["a"])

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_zero_total_weight_error(
        self,
        minimal_attribute_dict,
        uninitialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test that an exception is raised if the total sample weights are 0."""

        args = minimal_attribute_dict[self.transformer_name].copy()
        weight_column = "weight_column"
        args["weights_column"] = weight_column

        df = minimal_dataframe_lookup[self.transformer_name]
        uninitialized_transformer = uninitialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not uninitialized_transformer.polars_compatible and isinstance(
            df,
            pl.DataFrame,
        ):
            return

        df = nw.from_native(df)
        df = df.with_columns(nw.lit(0).alias("weight_column"))
        df = nw.to_native(df)

        transformer = uninitialized_transformer(**args)
        with pytest.raises(
            ValueError,
            match="total sample weights are not greater than 0",
        ):
            transformer.fit(df, df["a"])


class GenericTransformTests:
    """
    Generic tests for transformer.transform().
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    @pytest.mark.parametrize("non_df", [1, True, "a", [1, 2], {"a": 1}, None])
    def test_non_pd_type_error(
        self,
        non_df,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test that an error is raised in transform is X is not a pd/pl.DataFrame."""

        df = minimal_dataframe_lookup[self.transformer_name]
        x = initialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        x_fitted = x.fit(df, df["a"])

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: X should be a polars or pandas DataFrame/LazyFrame",
        ):
            x_fitted.transform(X=non_df)

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_no_rows_error(
        self,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test an error is raised if X has no rows."""

        df = minimal_dataframe_lookup[self.transformer_name]
        x = initialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        x = x.fit(df, df["a"])

        df = df.head(0)

        with pytest.raises(
            ValueError,
            match=re.escape(f"{self.transformer_name}: X has no rows; {df.shape}"),
        ):
            x.transform(df)

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=True,
    )
    def test_original_df_not_updated(
        self,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test that the original dataframe is not transformed when transform method used."""

        df = minimal_dataframe_lookup[self.transformer_name]
        x = initialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        original_df = copy.deepcopy(df)

        x = x.fit(df, df["a"])

        _ = x.transform(df)

        assert_frame_equal_dispatch(df, original_df)

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas"],
        indirect=True,
    )
    def test_pandas_index_not_updated(
        self,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test that the original (pandas) dataframe index is not transformed when transform method used."""

        df = minimal_dataframe_lookup[self.transformer_name]
        x = initialized_transformers[self.transformer_name]

        # update to abnormal index
        df.index = [2 * i for i in df.index]

        original_df = copy.deepcopy(df)

        x = x.fit(df, df["a"])

        _ = x.transform(df)

        assert_frame_equal_dispatch(df, original_df)


class DropOriginalTransformMixinTests:
    """
    Transform tests for transformers that take a "drop_original" argument
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=["minimal_dataframe_lookup"],
    )
    def test_original_columns_dropped_when_specified(
        self,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test transformer drops original columns when specified."""

        df = minimal_dataframe_lookup[self.transformer_name]

        x = initialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        x.drop_original = True

        x.fit(df)

        df_transformed = x.transform(df)
        remaining_cols = df_transformed.columns
        for col in x.columns:
            assert col not in remaining_cols, "original columns not dropped"

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=["minimal_dataframe_lookup"],
    )
    def test_original_columns_kept_when_specified(
        self,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test transformer keeps original columns when specified."""

        df = minimal_dataframe_lookup[self.transformer_name]

        x = initialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        x.drop_original = False

        x.fit(df)

        df_transformed = x.transform(df)
        remaining_cols = df_transformed.columns
        for col in x.columns:
            assert col in remaining_cols, "original columns not kept"

    @pytest.mark.parametrize(
        "minimal_dataframe_lookup",
        ["pandas", "polars"],
        indirect=["minimal_dataframe_lookup"],
    )
    def test_other_columns_not_modified(
        self,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test transformer does not modify unspecified columns."""

        df = minimal_dataframe_lookup[self.transformer_name]

        x = initialized_transformers[self.transformer_name]

        # skip polars test if not narwhalified
        if not x.polars_compatible and isinstance(df, pl.DataFrame):
            return

        other_columns = list(set(df.columns) - set(x.columns))
        x.drop_original = True

        x.fit(df)

        df_transformed = x.transform(df)

        assert_frame_equal_dispatch(df[other_columns], df_transformed[other_columns])


class ColumnsCheckTests:
    """
    Tests for columns_check method.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    def test_non_pd_df_error(
        self,
        initialized_transformers,
    ):
        """Test an error is raised if X is not passed as a pd.DataFrame."""

        x = initialized_transformers[self.transformer_name]

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: X should be a polars or pandas DataFrame/LazyFrame",
        ):
            x.columns_check(X=[1, 2, 3, 4, 5, 6])

    @pytest.mark.parametrize("non_list", [1, True, {"a": 1}, None, "True"])
    def test_columns_not_list_error(
        self,
        non_list,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test an error is raised if self.columns is not a list."""
        df = minimal_dataframe_lookup[self.transformer_name]

        x = initialized_transformers[self.transformer_name]

        x.columns = non_list

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: self.columns should be a list",
        ):
            x.columns_check(X=df)

    def test_columns_not_in_X_error(
        self,
        initialized_transformers,
        minimal_dataframe_lookup,
    ):
        """Test an error is raised if self.columns contains a value not in X."""
        df = minimal_dataframe_lookup[self.transformer_name]

        x = initialized_transformers[self.transformer_name]

        x.columns = ["a", "z"]

        with pytest.raises(ValueError):
            x.columns_check(X=df)


class CombineXYTests:
    """
    Tests for the BaseTransformer._combine_X_y method.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    @pytest.mark.parametrize("non_df", [1, True, "a", [1, 2], {"a": 1}, None])
    def test_X_not_DataFrame_error(
        self,
        non_df,
        initialized_transformers,
    ):
        """Test an exception is raised if X is not a pd.DataFrame."""

        x = initialized_transformers[self.transformer_name]

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: X should be a polars or pandas DataFrame/LazyFrame",
        ):
            x._combine_X_y(X=non_df, y=pd.Series([1, 2]))

    @pytest.mark.parametrize("non_series", [1, True, "a", [1, 2], {"a": 1}])
    def test_y_not_Series_error(
        self,
        non_series,
        initialized_transformers,
    ):
        """Test an exception is raised if y is not a pd.Series."""

        x = initialized_transformers[self.transformer_name]

        with pytest.raises(
            TypeError,
            match=f"{self.transformer_name}: y should be a polars or pandas Series",
        ):
            x._combine_X_y(X=pd.DataFrame({"a": [1, 2]}), y=non_series)

    def test_X_and_y_different_number_of_rows_error(
        self,
        initialized_transformers,
    ):
        """Test an exception is raised if X and y have different numbers of rows."""

        x = initialized_transformers[self.transformer_name]

        with pytest.raises(
            ValueError,
            match=re.escape(
                f"{self.transformer_name}: X and y have different numbers of rows (2 vs 1)",
            ),
        ):
            x._combine_X_y(X=pd.DataFrame({"a": [1, 2]}), y=pd.Series([2]))

    def test_output_same_indexes(
        self,
        initialized_transformers,
    ):
        """Test output is correct if X and y have the same index."""
        x = initialized_transformers[self.transformer_name]

        result = x._combine_X_y(
            X=pd.DataFrame({"a": [1, 2]}, index=[1, 2]),
            y=pd.Series([2, 4], index=[1, 2]),
        )

        expected_output = pd.DataFrame(
            {"a": [1, 2], "_temporary_response": [2, 4]},
            index=[1, 2],
        )

        pd.testing.assert_frame_equal(result, expected_output)


class OtherBaseBehaviourTests(
    ColumnsCheckTests,
    CombineXYTests,
):
    """
    Class to collect and hold tests for BaseTransformerBehaviour outside the three standard methods.
    Note this deliberately avoids starting with "Tests" so that the tests are not run on import.
    """

    def test_is_serialisable(self, initialized_transformers, tmp_path):
        path = tmp_path / "transformer.pkl"

        # serialise without raising error
        joblib.dump(initialized_transformers[self.transformer_name], path)



from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw
import pandas as pd

if TYPE_CHECKING:
    from narwhals.typing import FrameT

from tubular.base import BaseTransformer


class SetValueTransformer(BaseTransformer):
    """Transformer to set value of column(s) to a given value.

    This should be used if columns need to be set to a constant value.

    Parameters
    ----------
    columns: list or str
        Columns to set values.

    value : various
        Value to set.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework

    """

    polars_compatible = True

    def __init__(
        self,
        columns: str | list[str],
        value: type,
        copy: bool | None = None,
        **kwargs: dict[str, bool],
    ) -> None:
        self.value = value

        super().__init__(columns=columns, copy=copy, **kwargs)

    @nw.narwhalify
    def transform(self, X: FrameT) -> FrameT:
        """Set columns to value.

        Parameters
        ----------
        X : FrameT
            Data to apply mappings to.

        Returns
        -------
        X : FrameT
            Transformed input X with columns set to value.

        """
        X = nw.from_native(super().transform(X))

        set_value_expression = [nw.lit(self.value).alias(col) for col in self.columns]
        return X.with_columns(set_value_expression)


class ColumnDtypeSetter(BaseTransformer):
    """Transformer to set transform columns in a dataframe to a dtype.

    Parameters
    ----------
    columns : str or list
        Columns to set dtype. Must be set or transform will not run.

    dtype : type or string
        dtype object to set columns to or a string interpretable as one by pd.api.types.pandas_dtype
        e.g. float or 'float'

    Attributes
    ----------

    polars_compatible : bool
        class attribute, indicates whether transformer has been converted to polars/pandas agnostic narwhals framework
    """

    polars_compatible = False

    def __init__(
        self,
        columns: str | list[str],
        dtype: type | str,
        **kwargs: dict[str, bool],
    ) -> None:
        super().__init__(columns, **kwargs)

        self.__validate_dtype(dtype)

        self.dtype = dtype

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = super().transform(X)

        X[self.columns] = X[self.columns].astype(self.dtype)

        return X

    def __validate_dtype(self, dtype: str) -> None:
        """Check string is a valid dtype."""
        try:
            pd.api.types.pandas_dtype(dtype)
        except TypeError:
            msg = f"{self.classname()}: data type '{dtype}' not understood as a valid dtype"
            raise TypeError(msg) from None



import narwhals as nw
import numpy as np
import pytest

import tests.test_data as d
from tests import utils as u
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    GenericTransformTests,
    OtherBaseBehaviourTests,
    WeightColumnFitMixinTests,
    WeightColumnInitMixinTests,
)
from tests.imputers.test_BaseImputer import (
    GenericImputerTransformTests,
    GenericImputerTransformTestsWeight,
)
from tubular.imputers import MedianImputer


class TestInit(ColumnStrListInitTests, WeightColumnInitMixinTests):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MedianImputer"


class TestFit(WeightColumnFitMixinTests, GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MedianImputer"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_learnt_values(self, library):
        """Test that the impute values learnt during fit are expected."""
        df = d.create_df_3(library=library)

        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df)

        # replace 'a' with all null values to trigger warning
        df = df.with_columns(
            nw.new_series(
                name="d",
                values=[None] * len(df),
                backend=native_backend,
            ),
        )

        df = df.to_native()

        transformer = MedianImputer(columns=["a", "b", "c", "d"])

        transformer.fit(df)

        assert transformer.impute_values_ == {
            "a": df["a"].median(),
            "b": df["b"].median(),
            "c": df["c"].median(),
            "d": None,
        }, "impute_values_ attribute"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_learnt_values_weighted(self, library):
        """Test that the impute values learnt during fit are expected - when using weights."""
        df = d.create_df_9(library=library)

        df = nw.from_native(df)
        native_backend = nw.get_native_namespace(df)

        # replace 'a' with all null values to trigger warning
        df = df.with_columns(
            nw.new_series(
                name="d",
                values=[None] * len(df),
                backend=native_backend,
            ),
        )

        df = df.to_native()

        transformer = MedianImputer(columns=["a", "d"], weights_column="c")

        transformer.fit(df)

        assert transformer.impute_values_ == {
            "a": np.int64(4),
            "d": None,
        }, "impute_values_ attribute"

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_fit_not_changing_data(self, library):
        """Test fit does not change X."""
        df = d.create_df_1(library=library)

        transformer = MedianImputer(columns="a")

        transformer.fit(df)

        # Check whole dataframes
        u.assert_frame_equal_dispatch(
            d.create_df_1(library=library),
            df,
        )


class TestTransform(
    GenericImputerTransformTests,
    GenericImputerTransformTestsWeight,
    GenericTransformTests,
):
    """Tests for transformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MedianImputer"


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour behaviour outside the three standard methods.
    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "MedianImputer"



from copy import deepcopy

import narwhals as nw
import polars as pl
import pytest
from sklearn.exceptions import NotFittedError

import tests.test_data as d
from tests import utils as u
from tests.base_tests import (
    ColumnStrListInitTests,
    GenericFitTests,
    OtherBaseBehaviourTests,
)

# Categorical columns created under the same global string cache have the same underlying
# physical value when string values are equal.
# there is an efficiency cost, but not an issue for tests
pl.enable_string_cache()


class GenericImputerTransformTests:
    @pytest.fixture()
    def test_fit_df(self, request):
        library = request.param
        df_dict = {
            "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
            "b": ["a", "b", "c", "d", "e", "f", None],
            "c": ["a", "b", "c", "d", "e", "f", None],
        }

        return u.dataframe_init_dispatch(df_dict, library)

    @pytest.fixture()
    def expected_df_1(self, request):
        library = request.param
        df1_dict = {
            "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
            "b": ["a", "b", "c", "d", "e", "f", None],
            "c": ["a", "b", "c", "d", "e", "f", None],
        }

        df1 = u.dataframe_init_dispatch(df1_dict, library)

        narwhals_df = nw.from_native(df1)
        narwhals_df = narwhals_df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

        return narwhals_df.to_native()

    @pytest.fixture()
    def expected_df_2(self, request):
        library = request.param
        df2_dict = {
            "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, None],
            "b": ["a", "b", "c", "d", "e", "f", "g"],
            "c": ["a", "b", "c", "d", "e", "f", None],
        }
        df2 = u.dataframe_init_dispatch(df2_dict, library)
        narwhals_df = nw.from_native(df2)
        narwhals_df = narwhals_df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

        return narwhals_df.to_native()

    @pytest.fixture()
    def expected_df_3(self, request):
        library = request.param
        df3_dict = {
            "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, None],
            "b": ["a", "b", "c", "d", "e", "f", "g"],
            "c": ["a", "b", "c", "d", "e", "f", "f"],
        }

        df3 = u.dataframe_init_dispatch(dataframe_dict=df3_dict, library=library)

        narwhals_df = nw.from_native(df3)
        narwhals_df = narwhals_df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

        return narwhals_df.to_native()

    @pytest.fixture()
    def expected_df_4(self, request):
        library = request.param
        df4_dict = {
            "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, None],
            "b": ["a", "b", "c", "d", "e", "f", "z"],
            "c": ["a", "b", "c", "d", "e", "f", "z"],
        }

        df4 = u.dataframe_init_dispatch(dataframe_dict=df4_dict, library=library)

        narwhals_df = nw.from_native(df4)
        narwhals_df = narwhals_df.with_columns(nw.col("c").cast(nw.dtypes.Categorical))

        return narwhals_df.to_native()

    @pytest.mark.parametrize("test_fit_df", ["pandas", "polars"], indirect=True)
    def test_not_fitted_error_raised(self, test_fit_df, initialized_transformers):
        if initialized_transformers[self.transformer_name].FITS:
            with pytest.raises(NotFittedError):
                initialized_transformers[self.transformer_name].transform(test_fit_df)

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_impute_value_unchanged(self, library, initialized_transformers):
        """Test that self.impute_value is unchanged after transform."""
        df1 = d.create_df_1(library=library)
        transformer = initialized_transformers[self.transformer_name]

        impute_value = "g"
        transformer.impute_values_ = {"b": impute_value}

        if self.transformer_name == "ArbitraryImputer":
            transformer.impute_value = impute_value

        impute_values = deepcopy(transformer.impute_values_)

        transformer.transform(df1)

        assert (
            transformer.impute_values_ == impute_values
        ), "impute_values_ changed in transform"

    @pytest.mark.parametrize(
        ("library", "expected_df_1"),
        [("pandas", "pandas"), ("polars", "polars")],
        indirect=["expected_df_1"],
    )
    def test_expected_output_1(self, library, expected_df_1, initialized_transformers):
        """Test that transform is giving the expected output when applied to float column."""
        # Create the DataFrame using the library parameter
        df2 = d.create_df_2(library=library)

        # Initialize the transformer
        transformer = initialized_transformers[self.transformer_name]

        transformer.impute_values_ = {"a": 7}

        if self.transformer_name == "ArbitraryImputer":
            transformer.impute_value = 7

        transformer.columns = ["a"]

        # Transform the DataFrame
        df_transformed = transformer.transform(df2)

        # Check whole dataframes
        u.assert_frame_equal_dispatch(
            df_transformed,
            expected_df_1,
        )
        df2 = nw.from_native(df2)
        expected_df_1 = nw.from_native(expected_df_1)

        # Check outcomes for single rows
        # turn off type change errors to avoid having to type the single rows
        transformer.error_on_type_change = False
        for i in range(len(df2)):
            df_transformed_row = transformer.transform(df2[[i]].to_native())
            df_expected_row = expected_df_1[[i]].to_native()

            u.assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row,
            )

    @pytest.mark.parametrize(
        ("library", "expected_df_2"),
        [("pandas", "pandas"), ("polars", "polars")],
        indirect=["expected_df_2"],
    )
    def test_expected_output_2(self, library, expected_df_2, initialized_transformers):
        """Test that transform is giving the expected output when applied to object column."""
        # Create the DataFrame using the library parameter
        df2 = d.create_df_2(library=library)

        # Initialize the transformer
        transformer = initialized_transformers[self.transformer_name]

        impute_value = "g"
        transformer.impute_values_ = {"b": impute_value}

        if self.transformer_name == "ArbitraryImputer":
            transformer.impute_value = impute_value

        transformer.columns = ["b"]

        # Transform the DataFrame
        df_transformed = transformer.transform(df2)

        # Check whole dataframes
        u.assert_frame_equal_dispatch(
            df_transformed,
            expected_df_2,
        )
        df2 = nw.from_native(df2)
        expected_df_2 = nw.from_native(expected_df_2)

        # Check outcomes for single rows
        # turn off type change errors to avoid having to type the single rows
        transformer.error_on_type_change = False
        for i in range(len(df2)):
            df_transformed_row = transformer.transform(df2[[i]].to_native())
            df_expected_row = expected_df_2[[i]].to_native()

            u.assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row,
            )

    @pytest.mark.parametrize(
        ("library", "expected_df_3", "impute_values_dict"),
        [
            ("pandas", "pandas", {"b": "g", "c": "f"}),
            ("polars", "polars", {"b": "g", "c": "f"}),
        ],
        indirect=["expected_df_3"],
    )
    def test_expected_output_3(
        self,
        library,
        expected_df_3,
        initialized_transformers,
        impute_values_dict,
    ):
        """Test that transform is giving the expected output when applied to object and categorical columns."""
        # Create the DataFrame using the library parameter
        df2 = d.create_df_2(library=library)

        # Initialize the transformer
        transformer = initialized_transformers[self.transformer_name]

        transformer.impute_values_ = impute_values_dict

        if self.transformer_name == "ArbitraryImputer":
            transformer.impute_value = "f"

        transformer.columns = ["b", "c"]

        # Transform the DataFrame
        df_transformed = transformer.transform(df2)

        # Check whole dataframes
        u.assert_frame_equal_dispatch(
            df_transformed,
            expected_df_3,
        )
        df2 = nw.from_native(df2)
        expected_df_3 = nw.from_native(expected_df_3)

        # Check outcomes for single rows
        # turn off type change errors to avoid having to type the single rows
        transformer.error_on_type_change = False
        for i in range(len(df2)):
            df_transformed_row = transformer.transform(df2[[i]].to_native())
            df_expected_row = expected_df_3[[i]].to_native()

            u.assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row,
            )

    @pytest.mark.parametrize(
        "library",
        ["pandas", "polars"],
    )
    @pytest.mark.parametrize(
        ("column", "impute_value", "expected"),
        [
            ("a", False, [True, False, False]),
            ("b", 0, [1, 2, 0]),
        ],
    )
    def test_imputation_with_falsey_values(
        self,
        library,
        initialized_transformers,
        column,
        impute_value,
        expected,
    ):
        """Test that transform is giving the expected output when applied to object and categorical columns."""
        # Create the DataFrame using the library parameter
        df_dict = {
            "a": [True, False, None],
            "b": [1, 2, None],
        }

        df = u.dataframe_init_dispatch(dataframe_dict=df_dict, library=library)

        df = nw.from_native(df)

        # pandas bool+null cols default to Object, cast to bool
        df = df.with_columns(nw.col("a").cast(nw.Boolean))
        df = nw.to_native(df)

        # Initialize the transformer
        transformer = initialized_transformers[self.transformer_name]

        transformer.impute_values_ = {column: impute_value}

        if self.transformer_name == "ArbitraryImputer":
            transformer.impute_value = impute_value

        transformer.columns = [column]

        expected_df_dict = {
            column: expected,
        }

        expected_df = u.dataframe_init_dispatch(
            dataframe_dict=expected_df_dict,
            library=library,
        )

        # make sure types align with original
        expected_df = nw.from_native(expected_df)
        expected_df = expected_df.with_columns(
            nw.col(column).cast(nw.from_native(df)[column].dtype),
        )

        # Transform the DataFrame
        df_transformed = transformer.transform(df)

        u.assert_frame_equal_dispatch(
            df_transformed[[column]],
            expected_df.to_native()[[column]],
        )


class GenericImputerTransformTestsWeight:
    @pytest.fixture()
    def expected_df_weights(self, request):
        """Expected output for test_nulls_imputed_correctly_weights."""
        library = request.param
        df = d.create_df_9(library=library)

        df = nw.from_native(df)

        df = df.with_columns(df["b"].fill_null(4))

        return df.to_native()

    @pytest.mark.parametrize(
        ("library", "expected_df_weights"),
        [("pandas", "pandas"), ("polars", "polars")],
        indirect=["expected_df_weights"],
    )
    def test_nulls_imputed_correctly_weights(
        self,
        library,
        expected_df_weights,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test missing values are filled with the correct values - and unrelated columns are not changed
        (when weight is used).
        """
        # Create the DataFrame using the library parameter
        df = d.create_df_9(library=library)

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["weights_column"] = "c"

        transformer = uninitialized_transformers[self.transformer_name](**args)

        # Set the impute values dict directly rather than fitting x on df so test works with helpers
        impute_value = 4
        transformer.impute_values_ = {"b": impute_value}

        if self.transformer_name == "ArbitraryImputer":
            self.impute_value = impute_value

        df_transformed = transformer.transform(df)

        # Convert both DataFrames to a common format using Narwhals
        df_transformed_common = nw.from_native(df_transformed)
        expected_df_weights_common = nw.from_native(expected_df_weights)

        # Check outcomes for single rows
        for i in range(len(df_transformed_common)):
            df_transformed_row = df_transformed_common[[i]].to_native()
            df_expected_row = expected_df_weights_common[[i]].to_native()

            u.assert_frame_equal_dispatch(
                df_transformed_row,
                df_expected_row,
            )

        # Check whole dataframes
        u.assert_frame_equal_dispatch(
            df_transformed_common.to_native(),
            expected_df_weights_common.to_native(),
        )

    @pytest.mark.parametrize("library", ["pandas", "polars"])
    def test_learnt_values_not_modified_weights(
        self,
        library,
        minimal_attribute_dict,
        uninitialized_transformers,
    ):
        """Test that the impute_values_ from fit are not changed in transform - when using weights."""
        df = d.create_df_9(library=library)

        args = minimal_attribute_dict[self.transformer_name].copy()
        args["columns"] = ["a", "b"]
        args["weights_column"] = "c"

        transformer1 = uninitialized_transformers[self.transformer_name](**args)

        transformer1.fit(df)

        transformer2 = uninitialized_transformers[self.transformer_name](**args)

        transformer2.fit_transform(df)

        # Check if the impute_values_ are the same
        assert (
            transformer1.impute_values_ == transformer2.impute_values_
        ), f"Impute values changed in transform for {self.transformer_name}"


class TestInit(ColumnStrListInitTests):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseImputer"


class TestFit(GenericFitTests):
    """Generic tests for transformer.fit()"""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseImputer"


class TestTransform(GenericImputerTransformTests):
    """Tests for BaseImputer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseImputer"


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseImputer"



import datetime

import numpy as np
import pandas as pd
import pytest
import test_aide as ta

from tests.base_tests import (
    ColumnStrListInitTests,
    DropOriginalInitMixinTests,
    GenericTransformTests,
    NewColumnNameInitMixintests,
    OtherBaseBehaviourTests,
)
from tubular.dates import ToDatetimeTransformer


class TestInit(
    NewColumnNameInitMixintests,
    DropOriginalInitMixinTests,
    ColumnStrListInitTests,
):
    """Generic tests for transformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseDatetimeTransformer"

    def test_to_datetime_kwargs_type_error(self):
        """Test that an exception is raised if to_datetime_kwargs is not a dict."""
        with pytest.raises(
            TypeError,
            match=r"""ToDatetimeTransformer: to_datetime_kwargs should be a dict but got type \<class 'int'\>""",
        ):
            ToDatetimeTransformer(column="b", new_column_name="a", to_datetime_kwargs=1)

    def test_to_datetime_kwargs_key_type_error(self):
        """Test that an exception is raised if to_datetime_kwargs has keys which are not str."""
        with pytest.raises(
            TypeError,
            match=r"""ToDatetimeTransformer: unexpected type \(\<class 'int'\>\) for to_datetime_kwargs key in position 1, must be str""",
        ):
            ToDatetimeTransformer(
                new_column_name="a",
                column="b",
                to_datetime_kwargs={"a": 1, 2: "b"},
            )


class TestTransform(GenericTransformTests):
    """Tests for BaseDatetimeTransformer.transform."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseDatetimeTransformer"

    def expected_df_1():
        """Expected output for test_expected_output."""
        return pd.DataFrame(
            {
                "a": [1950, 1960, 2000, 2001, np.nan, 2010],
                "b": [1, 2, 3, 4, 5, np.nan],
                "a_Y": [
                    datetime.datetime(1950, 1, 1, tzinfo=datetime.timezone.utc),
                    datetime.datetime(1960, 1, 1, tzinfo=datetime.timezone.utc),
                    datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc),
                    datetime.datetime(2001, 1, 1, tzinfo=datetime.timezone.utc),
                    pd.NaT,
                    datetime.datetime(2010, 1, 1, tzinfo=datetime.timezone.utc),
                ],
                "b_m": [
                    datetime.datetime(1900, 1, 1, tzinfo=datetime.timezone.utc),
                    datetime.datetime(1900, 2, 1, tzinfo=datetime.timezone.utc),
                    datetime.datetime(1900, 3, 1, tzinfo=datetime.timezone.utc),
                    datetime.datetime(1900, 4, 1, tzinfo=datetime.timezone.utc),
                    datetime.datetime(1900, 5, 1, tzinfo=datetime.timezone.utc),
                    pd.NaT,
                ],
            },
        )

    def create_to_datetime_test_df():
        """Create DataFrame to be used in the ToDatetimeTransformer tests."""
        return pd.DataFrame(
            {"a": [1950, 1960, 2000, 2001, np.nan, 2010], "b": [1, 2, 3, 4, 5, np.nan]},
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(
            create_to_datetime_test_df(),
            expected_df_1(),
        ),
    )
    def test_expected_output(self, df, expected):
        """Test input data is transformed as expected."""
        to_dt_1 = ToDatetimeTransformer(
            column="a",
            new_column_name="a_Y",
            to_datetime_kwargs={"format": "%Y", "utc": datetime.timezone.utc},
        )

        to_dt_2 = ToDatetimeTransformer(
            column="b",
            new_column_name="b_m",
            to_datetime_kwargs={"format": "%m", "utc": datetime.timezone.utc},
        )

        df_transformed = to_dt_1.transform(df)
        df_transformed = to_dt_2.transform(df_transformed)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="ToDatetimeTransformer.transform output",
        )


class TestOtherBaseBehaviour(OtherBaseBehaviourTests):
    """
    Class to run tests for BaseTransformerBehaviour outside the three standard methods.

    May need to overwite specific tests in this class if the tested transformer modifies this behaviour.
    """

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "BaseDatetimeTransformer"



import numpy as np
import pandas as pd
import pytest
import test_aide as ta

import tests.test_data as d
from tests.numeric.test_BaseNumericTransformer import (
    BaseNumericTransformerInitTests,
    BaseNumericTransformerTransformTests,
)
from tubular.numeric import InteractionTransformer


class TestInit(BaseNumericTransformerInitTests):
    """Tests for InteractionTransformer.init()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "InteractionTransformer"

    def test_invalid_input_type_errors(self):
        """Test that an exceptions are raised for invalid input types."""

        with pytest.raises(
            TypeError,
            match=r"""InteractionTransformer: unexpected type \(\<class 'str'\>\) for min_degree, must be int""",
        ):
            InteractionTransformer(
                columns=["A", "B", "C"],
                min_degree="2",
                max_degree=2,
            )
        with pytest.raises(
            TypeError,
            match=r"""InteractionTransformer: unexpected type \(\<class 'str'\>\) for max_degree, must be int""",
        ):
            InteractionTransformer(
                columns=["A", "B", "C"],
                min_degree=2,
                max_degree="2",
            )

    def test_invalid_input_value_errors(self):
        """Test and exception is raised if degrees or columns provided are inconsistent."""
        with pytest.raises(
            ValueError,
            match=r"""InteractionTransformer: number of columns must be equal or greater than 2, got 1 column.""",
        ):
            InteractionTransformer(
                columns=["A"],
            )

        with pytest.raises(
            ValueError,
            match=r"""InteractionTransformer: min_degree must be equal or greater than 2, got 0""",
        ):
            InteractionTransformer(
                columns=["A", "B", "C"],
                min_degree=0,
                max_degree=2,
            )

        with pytest.raises(
            ValueError,
            match=r"""InteractionTransformer: max_degree must be equal or greater than min_degree""",
        ):
            InteractionTransformer(
                columns=["A", "B", "C"],
                min_degree=3,
                max_degree=2,
            )
        # NEW
        with pytest.raises(
            ValueError,
            match=r"""InteractionTransformer: max_degree must be equal or lower than number of columns""",
        ):
            InteractionTransformer(
                columns=["A", "B", "C"],
                min_degree=3,
                max_degree=4,
            )


class TestTransform(BaseNumericTransformerTransformTests):
    """Tests for InteractionTransformer.transform()."""

    @classmethod
    def setup_class(cls):
        cls.transformer_name = "InteractionTransformer"

    def expected_df_1():
        """Expected output of test_expected_output_default_assignment."""
        return pd.DataFrame(
            {
                "a": {0: 1.0, 1: 2.0, 2: 3.0, 3: 4.0, 4: 5.0, 5: 6.0, 6: np.nan},
                "b": {0: 1.0, 1: 2.0, 2: 3.0, 3: np.nan, 4: 7.0, 5: 8.0, 6: 9.0},
                "c": {0: np.nan, 1: 1.0, 2: 2.0, 3: 3.0, 4: -4.0, 5: -5.0, 6: -6.0},
                "a b": {0: 1.0, 1: 4.0, 2: 9.0, 3: np.nan, 4: 35.0, 5: 48.0, 6: np.nan},
                "a c": {
                    0: np.nan,
                    1: 2.0,
                    2: 6.0,
                    3: 12.0,
                    4: -20.0,
                    5: -30.0,
                    6: np.nan,
                },
                "b c": {
                    0: np.nan,
                    1: 2.0,
                    2: 6.0,
                    3: np.nan,
                    4: -28.0,
                    5: -40.0,
                    6: -54.0,
                },
            },
        )

    def expected_df_2():
        """Expected output of test_expected_output_multiple_columns_assignment."""
        return pd.DataFrame(
            {
                "a": {0: 1.0, 1: 2.0, 2: 3.0, 3: 4.0, 4: 5.0, 5: 6.0, 6: np.nan},
                "b": {0: 1.0, 1: 2.0, 2: 3.0, 3: np.nan, 4: 7.0, 5: 8.0, 6: 9.0},
                "c": {0: np.nan, 1: 1.0, 2: 2.0, 3: 3.0, 4: -4.0, 5: -5.0, 6: -6.0},
                "a b": {0: 1.0, 1: 4.0, 2: 9.0, 3: np.nan, 4: 35.0, 5: 48.0, 6: np.nan},
            },
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_3(), expected_df_1()),
    )
    def test_expected_output_default_assignment(self, df, expected):
        """Test default values and multiple columns assignment from transform gives expected results."""
        x = InteractionTransformer(columns=["a", "b", "c"])

        df_transformed = x.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="InteractionTransformer default values",
        )

    @pytest.mark.parametrize(
        ("df", "expected"),
        ta.pandas.adjusted_dataframe_params(d.create_df_3(), expected_df_2()),
    )
    def test_expected_output_multiple_columns_assignment(self, df, expected):
        """Test a multiple columns assignment from transform gives expected results."""
        x = InteractionTransformer(columns=["a", "b"])

        df_transformed = x.transform(df)

        ta.equality.assert_equal_dispatch(
            expected=expected,
            actual=df_transformed,
            msg="InteractionTransformer one single column values",
        )
